from random import choice
from string import ascii_uppercase, digits
import requests

DOMAINS = [
    'getnada.com',
    'amail.club',
    'banit.club',
    'cars2.club',
    'cmail.club',
    'banit.me',
    'duck2.club',
    'nada.email',
    'nada.ltd',
    'wmail.club'
]

GET_INBOX = 'https://getnada.com/api/v1/inboxes/{}'
GET_MESSAGE = 'https://getnada.com/api/v1/messages/{}'


class EmailResult:
    def __init__(self, guid: str, subject: str, sender: str, datetime: str, body: str = None):
        self.body = body
        self.datetime = datetime
        self.sender = sender
        self.subject = subject
        self.guid = guid


class Nada:
    def __init__(self, name=None, domain=None):
        self.name = (name or Nada.id_generator()).lower()
        self.domain = domain or choice(DOMAINS)

    def __str__(self):
        return '{}@{}'.format(self.name, self.domain)

    def get_email_list(self):
        result = []
        try:
            response = requests.get(GET_INBOX.format(self))
            data = response.json()['msgs']
            result = [EmailResult(x['uid'], x['s'], x['fe'], x['rf']) for x in data]
        except Exception:
            pass

        return result

    @staticmethod
    def get_email(uid):
        result: EmailResult = None
        try:
            response = requests.get(GET_MESSAGE.format(uid))
            data = response.json()  # [ru, d, f, text, fe, html, s, response, at, rf, ib, uid]
            result = EmailResult(data['uid'], data['s'], data['fe'], data['rf'], data['text'] or data['html'])
        except Exception:
            pass
        return result

    @staticmethod
    def id_generator(size=6, chars=ascii_uppercase + digits):
        return ''.join(choice(chars) for _ in range(size))



# pylint: disable=too-many-instance-attributes
"""
Container for Response objects used in cs18-api-client
"""
import abc

from datetime import datetime
from typing import List, Dict

from gateways.common.cs18_api_classes import AccessLink, CloudComputeService, CloudAccountCostResponse
from gateways.common.cs18_api_errors import SandboxError, Error


class AccountStatusUserDetails:
    def __init__(self, user_details: {}):
        self.last_name = user_details["last_name"]
        self.first_name = user_details["first_name"]
        self.email = user_details["email"]
        self.permitted_spaces = user_details["permitted_spaces"]
        self.account_permissions = user_details["account_permissions"]


class AccountStatusResponse:
    def __init__(
            self,
            onboarding_status: str,
            license_expiration_date: datetime,
            ignore_expiration_date: bool,
            suspended_reason: str,
            salesforce_account_id: str,
            plan_id: str,
            user_details: AccountStatusUserDetails,
    ):
        self.user_details = user_details  # type: AccountStatusUserDetails
        self.salesforce_account_id = salesforce_account_id
        self.suspended_reason = suspended_reason
        self.license_expiration_date = license_expiration_date
        self.ignore_expiration_date = ignore_expiration_date
        self.plan_id = plan_id
        self.onboarding_status = onboarding_status


class TokenResponse:
    def __init__(
            self, access_token: str, refresh_token: str, token_type: str, expires_in: int
    ):
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.token_type = token_type
        self.expires_in = expires_in


class HyperlinkResponse:
    def __init__(self, rel: str, href: str, method: str):
        self.rel = rel
        self.href = href
        self.method = method


class CatalogApplicationResponse:
    def __init__(self, name: str, version: str):
        self.name = name
        self.version = version


class CatalogForGetResponse:
    def __init__(
            self,
            name: str,
            description: str,
            url: str,
            last_modified: str,
            modified_by: str,
            clouds: List[str],
            applications: List[CatalogApplicationResponse],
            links: List[HyperlinkResponse],
            artifacts: Dict[str, str],
            errors: [],
    ):
        self.errors = errors
        self.links = links
        self.url = url
        self.description = description
        self.name = name
        self.last_modified = last_modified
        self.modified_by = modified_by
        self.clouds = clouds
        self.applications = applications
        self.artifacts = artifacts


class CatalogForGetAllResponse:
    def __init__(
            self,
            name: str,
            description: str,
            url: str,
            last_modified: str,
            modified_by: str,
            clouds: List[str],
            links: List[HyperlinkResponse],
            artifacts: Dict[str, str],
            errors: [],
    ):
        self.links = links
        self.url = url
        self.description = description
        self.name = name
        self.last_modified = last_modified
        self.modified_by = modified_by
        self.clouds = clouds
        self.artifacts = artifacts
        self.errors = errors


class BlueprintApplicationResponse:
    def __init__(self, name: str, version: str):
        self.name = name
        self.version = version


class BlueprintServiceResponse:
    def __init__(self, name: str, service_type: str):
        self.name = name
        self.type = service_type


class BlueprintInputResponse:
    def __init__(
            self, name: str, default_value: str, display_style: str, description: str, optional: bool
    ):
        self.name = name
        self.default_value = default_value
        self.display_style = display_style
        self.optional = optional
        self.description = description


class BlueprintRegionResponse:
    def __init__(self, region_id: str, name: str):
        self.name = name
        self.region_id = region_id


class BlueprintCloudResponse:
    def __init__(
            self, provider: str, cloud_account_name: str, region: BlueprintRegionResponse
    ):
        self.provider = provider
        self.cloud_account_name = cloud_account_name
        self.region = region


class ArtifactRepoCreatedUser:
    def __init__(self, email: str, first_name: str, last_name: str):
        self.last_name = last_name
        self.first_name = first_name
        self.email = email


class ArtifactRepoResponse:
    def __init__(
            self,
            repo_type: str,
            created_date: datetime,
            location: str,
            created_by: ArtifactRepoCreatedUser,
    ):
        self.created_by = created_by
        self.location = location
        self.created_date = created_date
        self.type = repo_type


class BlueprintResponse:
    def __init__(
            self,
            name: str,
            description: str,
            url: str,
            enabled: bool,
            last_modified: str,
            modified_by: str,
            clouds: List[BlueprintCloudResponse],
            applications: List[BlueprintApplicationResponse],
            links: List[HyperlinkResponse],
            artifacts: Dict[str, str],
            inputs: List[BlueprintInputResponse],
            errors: List[Error],
            is_sample,
            services: List[BlueprintServiceResponse]
    ):
        self.errors = errors
        self.is_sample = is_sample
        self.links = links
        self.enabled = enabled
        self.url = url
        self.description = description
        self.name = name
        self.last_modified = last_modified
        self.modified_by = modified_by
        self.clouds = clouds
        self.applications = applications
        self.artifacts = artifacts
        self.inputs = inputs
        self.services = services


class SourceControlDetailsResponse:
    def __init__(
            self,
            branch_name: str,
            latest_commit_id: str,
            changed_by: str,
            changed_date: datetime,
    ):
        self.branch_name = branch_name
        self.latest_commit_id = latest_commit_id
        self.changed_by = changed_by
        self.changed_date = changed_date


class BlueprintFileResponse:
    def __init__(
            self,
            url: str,
            kind: str,
            content: str,
            source_control_details: SourceControlDetailsResponse,
            icon: str,
    ):
        self.url = url
        self.kind = kind
        self.content = content
        self.source_control_details = source_control_details
        self.icon = icon


class CreateSandboxResponse:
    def __init__(self, sandbox_id: str):
        self.sandbox_id = sandbox_id


class CreateProductionResponse:
    def __init__(self, production_id: str):
        self.production_id = production_id


class AppInstanceResponse:
    def __init__(
            self,
            private_ip: str,
            infra_id: str,
            deployment_status: str,
            access_links: List[AccessLink],
            links: List[HyperlinkResponse],
            instance_type: str = None,
            memory_limit: str = None,
            cpu_limit: str = None,
    ):
        self.memory_limit = memory_limit
        self.cpu_limit = cpu_limit
        self.instance_type = instance_type
        self.links = links
        self.access_links = access_links
        self.private_ip = private_ip
        self.deployment_status = deployment_status
        self.infra_id = infra_id


class CloudRegionResponse:
    def __init__(self, region_id: str, region_name: str):
        self.region_id = region_id
        self.name = region_name


class CloudResponse:
    def __init__(
            self,
            cloud_account_name: str,
            region_id: str,
            region_name: str = None
    ):
        self.cloud_account_name = cloud_account_name
        self.region = CloudRegionResponse(region_id=region_id, region_name=region_name)


class ApplicationResponseBase:
    def __init__(self, name: str, status: str, cloud: CloudResponse):
        self.name = name
        self.status = status
        self.cloud = cloud


class ApplicationResponseLean(ApplicationResponseBase):
    def __init__(self, name: str, status: str, cloud: CloudResponse):
        super().__init__(name=name, status=status, cloud=cloud)


class ApplicationResponse(ApplicationResponseBase):
    def __init__(
            self,
            name: str,
            shortcuts: List[str],
            status: str,
            instances: List[AppInstanceResponse],
            private_address: str,
            public_address: str,
            internal_ports: str,
            external_ports: str,
            internal_dns: str,
            image_name: str,
            icon: str,
            cloud: CloudResponse = None,
    ):
        super().__init__(name=name, status=status, cloud=cloud)
        self.image_name = image_name
        self.internal_dns = internal_dns
        self.external_ports = external_ports
        self.internal_ports = internal_ports
        self.public_address = public_address
        self.private_address = private_address
        self.shortcuts = shortcuts
        self.instances = instances
        self.icon = icon


class SandboxOwnerResponse:
    def __init__(self, first_name: str, last_name: str, email: str, join_date: str):
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.join_date = join_date


class LaunchingProgressStep:
    def __init__(self, total: int, succeeded: int, failed: int, status: str):
        self.failed = failed
        self.succeeded = succeeded
        self.total = total
        self.status = status


class LaunchingProgress:
    def __init__(self,
                 creating_infrastructure: LaunchingProgressStep,
                 preparing_artifacts: LaunchingProgressStep,
                 deploying_applications: LaunchingProgressStep,
                 verifying_environment: LaunchingProgressStep):
        self.verifying_environment = verifying_environment
        self.deploying_applications = deploying_applications
        self.preparing_artifacts = preparing_artifacts
        self.creating_infrastructure = creating_infrastructure


class SandboxResponseBase:
    def __init__(self,
                 sandbox_status: str,
                 create_time: str,
                 start_time: str,
                 status_details: str,
                 status_error_description: str,
                 links: List[HyperlinkResponse],
                 blueprint: str,
                 automation: bool,
                 owner: SandboxOwnerResponse,
                 sandbox_id: str,
                 errors: List[SandboxError],
                 scheduled_end_time: datetime,
                 end_time: datetime,
                 debugging_service: str):
        self.sandbox_status = sandbox_status
        self.create_time = create_time
        self.start_time = start_time
        self.status_details = status_details
        self.status_error_description = status_error_description
        self.links = links
        self.sandbox_id = sandbox_id
        self.blueprint = blueprint
        self.automation = automation
        self.owner = owner
        self.errors = errors
        self.scheduled_end_time = scheduled_end_time
        self.end_time = end_time
        self.debugging_service = debugging_service


class InputResponse:
    def __init__(self, name: str, value: str, display_style: str, description: str, optional: bool):
        self.description = description
        self.display_style = display_style
        self.optional = optional
        self.value = value
        self.name = name


class ServiceResponse:
    def __init__(self,
                 name: str,
                 service_type: str,
                 status: str):
        self.name = name
        self.type = service_type
        self.status = status


class SandboxResponse(SandboxResponseBase):
    def __init__(self,
                 sandbox_id: str,
                 blueprint: str,
                 automation: bool,
                 applications: List[ApplicationResponse],
                 services: List[ServiceResponse],
                 inputs: List[InputResponse],
                 artifacts: Dict[str, str],
                 start_time: str,
                 create_time: str,
                 status_details: str,
                 links: List[HyperlinkResponse],
                 sandbox_status: str,
                 status_error_description: str,
                 owner: SandboxOwnerResponse,
                 errors: List[SandboxError],
                 scheduled_end_time: datetime,
                 end_time: datetime,
                 debugging_service: str,
                 launching_progress: LaunchingProgress):
        super().__init__(sandbox_status,
                         create_time,
                         start_time,
                         status_details,
                         status_error_description,
                         links,
                         blueprint,
                         automation,
                         owner,
                         sandbox_id,
                         errors,
                         scheduled_end_time,
                         end_time,
                         debugging_service)
        self.artifacts = artifacts
        self.inputs = inputs
        self.applications = applications
        self.services = services
        self.launching_progress = launching_progress


class SandboxResponseLean(SandboxResponseBase):
    def __init__(self,
                 sandbox_id: str,
                 blueprint: str,
                 automation: bool,
                 applications: List[ApplicationResponseLean],
                 services: List[ServiceResponse],
                 start_time: str,
                 create_time: str,
                 status_details: str,
                 status_error_description: str,
                 links: List[HyperlinkResponse],
                 sandbox_status: str,
                 owner: SandboxOwnerResponse,
                 name: str,
                 errors: List[SandboxError],
                 scheduled_end_time: datetime,
                 end_time: datetime,
                 debugging_service: str):
        super().__init__(sandbox_status,
                         create_time,
                         start_time,
                         status_details,
                         status_error_description,
                         links,
                         blueprint,
                         automation,
                         owner,
                         sandbox_id,
                         errors,
                         scheduled_end_time,
                         end_time,
                         debugging_service)
        self.end_time = end_time
        self.applications = applications
        self.name = name
        self.services = services


class ProductionResponseLean:
    def __init__(
            self,
            production_id: str,
            name: str,
            blueprint_description: str,
            blue_status: str,
            green_status: str,
            update_strategy: str,
            update_status: str,
            create_time: datetime,
            deployment_start_time: datetime,
            blue_applications: List[ApplicationResponseLean],
            blue_services: List[ServiceResponse]):
        self.production_id = production_id
        self.name = name
        self.blueprint_description = blueprint_description
        self.blue_status = blue_status
        self.green_status = green_status
        self.update_strategy = update_strategy
        self.update_status = update_status
        self.create_time = create_time
        self.deployment_start_time = deployment_start_time
        self.blue_applications = blue_applications
        self.blue_services = blue_services


class ProductionResponse(metaclass=abc.ABCMeta):
    def __init__(
            self,
            production_id: str,
            sandbox_id: str,
            name: str,
            owner: SandboxOwnerResponse,
            blueprint_name: str,
            blueprint_description: str,
            applications: List[ApplicationResponse],
            services: List[ServiceResponse],
            artifacts: Dict[str, str],
            inputs: List[InputResponse],
            create_time: datetime,
            deployment_start_time: datetime,
            build: str,
            production_status: str,
            update_status: str,
            status_details: str,
            status_error_description: str,
            errors: List[SandboxError],
            links: List[HyperlinkResponse],
            debugging_service: str,
            launching_progress: LaunchingProgress):
        self.production_id = production_id
        self.sandbox_id = sandbox_id
        self.name = name
        self.owner = owner
        self.blueprint_name = blueprint_name
        self.blueprint_description = blueprint_description
        self.applications = applications
        self.artifacts = artifacts
        self.inputs = inputs
        self.create_time = create_time
        self.deployment_start_time = deployment_start_time
        self.build = build
        self.production_status = production_status
        self.update_status = update_status
        self.status_details = status_details
        self.status_error_description = status_error_description
        self.errors = errors
        self.links = links
        self.services = services
        self.debugging_service = debugging_service
        self.launching_progress = launching_progress


class ProductionBlueResponse(ProductionResponse):
    def __init__(self,
                 production_id: str,
                 sandbox_id: str,
                 name: str,
                 owner: SandboxOwnerResponse,
                 blueprint_name: str,
                 blueprint_description: str,
                 applications: List[ApplicationResponse],
                 services: List[ServiceResponse],
                 artifacts: Dict[str, str],
                 inputs: List[InputResponse],
                 create_time: datetime,
                 deployment_start_time: datetime,
                 build: str,
                 production_status: str,
                 update_status: str,
                 status_details: str,
                 status_error_description: str,
                 errors: List[SandboxError],
                 links: List[HyperlinkResponse],
                 has_green: bool,
                 debugging_service: str,
                 launching_progress: LaunchingProgress):
        super().__init__(
            production_id=production_id,
            sandbox_id=sandbox_id,
            name=name,
            owner=owner,
            blueprint_name=blueprint_name,
            blueprint_description=blueprint_description,
            applications=applications,
            services=services,
            artifacts=artifacts,
            inputs=inputs,
            create_time=create_time,
            deployment_start_time=deployment_start_time,
            build=build,
            production_status=production_status,
            update_status=update_status,
            status_details=status_details,
            status_error_description=status_error_description,
            errors=errors,
            links=links,
            debugging_service=debugging_service,
            launching_progress=launching_progress)
        self.has_green = has_green


class ProductionGreenResponse(ProductionResponse):
    def __init__(self,
                 production_id: str,
                 sandbox_id: str,
                 name: str,
                 owner: SandboxOwnerResponse,
                 blueprint_name: str,
                 blueprint_description: str,
                 applications: List[ApplicationResponse],
                 services: List[ServiceResponse],
                 artifacts: Dict[str, str],
                 inputs: List[InputResponse],
                 create_time: datetime,
                 deployment_start_time: datetime,
                 build: str,
                 production_status: str,
                 update_status: str,
                 status_details: str,
                 status_error_description: str,
                 errors: List[SandboxError],
                 links: List[HyperlinkResponse],
                 exposure: int,
                 debugging_service: str,
                 launching_progress: LaunchingProgress):
        super().__init__(
            production_id=production_id,
            sandbox_id=sandbox_id,
            name=name,
            owner=owner,
            blueprint_name=blueprint_name,
            blueprint_description=blueprint_description,
            applications=applications,
            services=services,
            artifacts=artifacts,
            inputs=inputs,
            create_time=create_time,
            deployment_start_time=deployment_start_time,
            build=build,
            production_status=production_status,
            update_status=update_status,
            status_details=status_details,
            status_error_description=status_error_description,
            errors=errors,
            links=links,
            debugging_service=debugging_service,
            launching_progress=launching_progress)
        self.exposure = exposure


class UserPermittedToSpaceResponse:
    def __init__(
            self,
            first_name: str,
            last_name: str,
            email: str,
            join_date: str,
            space_role: str,
            account_role: str,
            has_access_to_all_spaces: bool):
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.join_date = join_date
        self.space_role = space_role
        self.account_role = account_role
        self.has_access_to_all_spaces = has_access_to_all_spaces


class CloudAccountResponse:
    def __init__(
            self,
            name: str,
            provider: str,
            created_date: datetime,
            created_by: str,
            spaces: List[str],
            compute_services: List[CloudComputeService] = None,
            cost: CloudAccountCostResponse = None,
            errors: List[Error] = None
    ):
        self.name = name
        self.provider = provider
        self.created_date = created_date
        self.created_by = created_by
        self.spaces = spaces
        self.compute_services = compute_services
        self.cost = cost
        self.errors = errors


class CloudAccountSubnetsResponse:
    def __init__(self, gateway_subnet: str,
                 management_subnet: str,
                 application_subnets: List[str]):
        self.gateway_subnet = gateway_subnet
        self.management_subnet = management_subnet
        self.application_subnets = application_subnets


class CloudAccountInfraSetResponse:
    def __init__(self,
                 region: str,
                 virtual_network: str,
                 subnets: CloudAccountSubnetsResponse):
        self.region = region
        self.virtual_network = virtual_network
        self.subnets = subnets


class CloudAccountConnectivityResponse:
    def __init__(self,
                 existing_infra: bool,
                 existing_infra_sets: List[CloudAccountInfraSetResponse]):
        self.existing_infra = existing_infra
        self.existing_infra_sets = existing_infra_sets


class CloudAccountInSpaceResponse:
    def __init__(
            self,
            name: str,
            provider: str,
            created_date: datetime,
            created_by: str,
            internet_facing: bool,
            connectivity: CloudAccountConnectivityResponse,
            compute_services: List[CloudComputeService] = None,
            cost: CloudAccountCostResponse = None,
            errors: List[Error] = None,
    ):
        self.name = name
        self.provider = provider
        self.created_date = created_date
        self.created_by = created_by
        self.compute_services = compute_services
        self.cost = cost
        self.errors = errors
        self.internet_facing = internet_facing
        self.connectivity = connectivity


class UserInvitationResponse:
    def __init__(
            self,
            email: str,
            send_date: str,
            account_role: str,
            space_name: str,
            space_role: str,
    ):
        self.email = email
        self.send_date = send_date
        self.account_role = account_role
        self.space_name = space_name
        self.space_role = space_role


class RepositoryResponse:
    def __init__(self, repository_url: str, errors: List[Error]):
        self.errors = errors
        self.repository_url = repository_url


class SpaceAdmin:
    def __init__(self, first_name: str, last_name: str, account_role: str):
        self.last_name = last_name
        self.first_name = first_name
        self.account_role = account_role


class SpaceCloudAccount:
    def __init__(self, name: str, cloud_account_type: str):
        self.type = cloud_account_type
        self.name = name


class GetSpaceResponse:
    def __init__(self, space: {}):
        self.name = space["name"]


class GetSpacesResponse:
    def __init__(
            self,
            name: str,
            admins: List[SpaceAdmin],
            users_count: int,
            cloud_accounts: List[SpaceCloudAccount],
    ):
        self.users_count = users_count
        self.cloud_accounts = cloud_accounts
        self.admins = admins
        self.name = name


class RoleListItemResponse:
    def __init__(self, name: str):
        self.name = name


class UserForAllUsersResponse:
    def __init__(self,
                 email: str,
                 first_name: str,
                 last_name: str,
                 join_date: str,
                 account_role: str,
                 has_access_to_all_spaces: bool,
                 member_in_spaces: List[str]):
        self.email = email
        self.first_name = first_name
        self.last_name = last_name
        self.join_date = join_date
        self.account_role = account_role
        self.has_access_to_all_spaces = has_access_to_all_spaces
        self.member_in_spaces = member_in_spaces


class RegionResponse:
    def __init__(
            self,
            region_id: str,
            region_name: str
    ):
        self.region_id = region_id
        self.region_name = region_name


class GetImagesResponse:
    def __init__(self, image_id: str, image_name: str):
        self.image_id = image_id
        self.image_name = image_name


class VerifyCloudProviderResourceResponse:
    def __init__(self,
                 resource_type: str,
                 resource_name: str,
                 resource_id: str = None,
                 resource_status: str = None,
                 resource_valid: bool = False):
        self.resource_type = resource_type
        self.resource_name = resource_name
        self.resource_id = resource_id
        self.resource_status = resource_status
        self.resource_valid = resource_valid


class VerifyCloudProviderResponse:
    def __init__(self, success: bool,
                 resources_summary: List[VerifyCloudProviderResourceResponse] = None):
        self.success = success
        self.resources_summary = resources_summary



from gateways.apis.api_base_class import ApiBase


class ApiAccount(ApiBase):
    def account_register(self):
        return self.build_route("accounts/register")

    def contact_me(self):
        return self.build_route("accounts/contactme")

    def account_login(self, account: str):
        return self.build_route(f"accounts/{account}/login")

    def account_details(self, account: str):
        return self.build_route(f"accounts/{account}/details")

    def forgot_password(self):
        return self.build_route("accounts/forgotpassword")

    def reset_password(self):
        return self.build_route("accounts/resetpassword")

    def invitations(self):
        return self.build_route("accounts/invitations")

    def account_signup_by_secret(self, secret: str):
        return self.build_route(f"accounts/signup/{secret}")

    def account_signup(self):
        return self.build_route("accounts/signup")

    def account_invitations(self, space: str = None):
        return self.build_route(
            "accounts/invitations" + (f"?space_name={space}" if space else "")
        )

    def account_users(self):
        return self.build_route("accounts/users")

    def account_user(self, user_email: str):
        return self.build_route(f"accounts/users/{user_email}")

    def account_user_account_role_value(self, user_email: str, account_role: str):
        return self.build_route(
            f"accounts/users/{user_email}/account_role?value={account_role}"
        )

    def account_user_account_role(self, user_email: str):
        return self.build_route(f"accounts/users/{user_email}/account_role")

    def delete_account(self, account_name: str) -> str:
        return self.build_route(f"accounts/{account_name}")

    def set_account_extra_details(self, account_name: str) -> str:
        return self.build_route(f"accounts/{account_name}/details")



import json
from datetime import datetime
from typing import List, Optional
from urllib.parse import quote

import time
from isodate import duration_isoformat

from gateways.apis.api_route_base import ApiRoot
from gateways.common.cs18_api_classes import BlueprintRepositoryDetails, AccessLink
from gateways.common.cs18_api_converters import Converters
from gateways.common.cs18_api_errors import UnauthorizedException, SandboxNotFound, \
    SandboxEndingFailed
from gateways.common.cs18_api_requests import (
    CreateInvitationsRequest,
    UpdateSpaceRequest,
    ProductionEnvironment, DebuggingServiceValue, SpaceCloudAccountInfraSettingsRequest)
from gateways.common.cs18_api_responses import (AccountStatusResponse, BlueprintFileResponse, BlueprintResponse,
                                                CatalogForGetAllResponse, CatalogForGetResponse, CloudAccountResponse,
                                                CreateProductionResponse, CreateSandboxResponse, GetSpaceResponse,
                                                GetSpacesResponse, ProductionBlueResponse, ProductionGreenResponse,
                                                ProductionResponseLean, RepositoryResponse, RoleListItemResponse,
                                                SandboxResponse, SandboxResponseLean, TokenResponse,
                                                UserInvitationResponse, UserPermittedToSpaceResponse,
                                                UserForAllUsersResponse, CloudAccountInSpaceResponse, RegionResponse,
                                                VerifyCloudProviderResponse)
from gateways.n_session import NSession
from gateways.the_gateway import TheGateway
from gateways.utils import GatewayUtils, Utils


class Colony:
    # pylint: disable=too-many-public-methods
    def __init__(self, account: str, **kwargs):
        """
        :param account: account name
        :param Keyword Arguments:
            * access_token: either access token, or long token
            * email: account email
            * password: account password
            * host: address of Colony host machine, including port. Example: 'http://192.168.30.22:5050'
            * space: current space to be used
            * api_version: api version to be used
        """
        # raising an error when not provided valid credentials
        if (not kwargs.get("access_token")) and not (kwargs.get("email") and kwargs.get("password")):
            credentials = {
                "access_token": kwargs.get("access_token"),
                "email": kwargs.get("email"),
                "password": kwargs.get("password")
            }
            raise UnauthorizedException(credentials)

        gateway = TheGateway(provider=kwargs.get("provider"), host=kwargs.get("host"))
        self.session = NSession()
        self.api_address = GatewayUtils.get_cs18_api_address(
            provider=kwargs.get("provider"), host=kwargs.get("host")
        ).lower()

        self.refresh_token = \
            kwargs.get("access_token") if kwargs.get("access_token") else \
                gateway.account_login(
                    account=account,
                    email=kwargs.get("email"),
                    password=kwargs.get("password")
                ).refresh_token

        # use either provided access_token, or login with provided credentials
        self.access_token = \
            kwargs.get("access_token") if kwargs.get("access_token") else \
                gateway.account_login(
                    account=account,
                    email=kwargs.get("email"),
                    password=kwargs.get("password")
                ).access_token

        self.session.add_header(
            "Authorization",
            "Bearer {}".format(self.access_token)
        )
        self.space = kwargs.get("space", "Trial")
        self._api_version = kwargs.get("api_version")
        self._api_root = ApiRoot(
            api_address=self.api_address, space=self.space, version=self._api_version
        )

    def change_space(self, space: str):
        self.space = space
        self._api_root = ApiRoot(
            api_address=self.api_address, space=self.space, version=self._api_version
        )
        return self

    def get_spaces(self) -> List[GetSpacesResponse]:
        method_url = self._api_root.spaces.spaces()
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        spaces_json = json.loads(resp.text)
        return [Converters.create_spaces_response(space) for space in spaces_json]

    def get_account_status(self) -> AccountStatusResponse:
        method_url = self._api_root.settings.status()
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        return Converters.create_account_status_response(json.loads(resp.text))

    def get_about(self):
        method_url = self._api_root.about.api_about()
        resp = self.session.get(url=method_url)
        json_response = json.loads(resp.text)
        return json_response

    def remove_compute_service(self, cloud_account: str, compute_service: str):
        method_url = self._api_root.settings.compute_services()
        body = {"compute_service_name": f'{cloud_account}/{compute_service}'}
        res = self.session.delete(url=method_url, json=body)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def get_cloud_accounts_under_account(self) -> List[CloudAccountResponse]:
        method_url = self._api_root.settings.cloud_accounts()
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        json_response = json.loads(resp.text)
        cloud_accounts = [
            Converters.create_cloud_account_response(x) for x in json_response
        ]
        return cloud_accounts

    def register_azure_account(self, with_tags: bool = False, **kwargs):
        """
        :param with_tags: bool
        :param kwargs:  name: str
                        subscription_id: str,
                        tenant_id: str,
                        application_id: str,
                        application_secret: str,
                        management_resource_group: str,
                        external_key: [str]
        """
        params = {"add_tags": with_tags}
        method_url = self._api_root.settings.azure_cloud_accounts()
        request = {
            "name": kwargs.get("name"),
            "subscription_id": kwargs.get("subscription_id"),
            "tenant_id": kwargs.get("tenant_id"),
            "application_id": kwargs.get("application_id"),
            "application_secret": kwargs.get("application_secret"),
            "management_resource_group": kwargs.get("management_resource_group"),
        }
        if kwargs.get("external_key"):
            request["id"] = kwargs.get("external_key")

        resp = self.session.post(url=method_url, json=request, params=params)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def register_aws_account(
            self,
            arn_role: str,
            external_id: str,
            name: str = None,
            external_key: str = None,
    ):
        """ Register AWS
        :param arn_role: IAM role ARN required for assume role to the customer cloud account
        :param external_id: External ID required for assume role to the customer cloud account
        :param name: Cloud account given name. If not passed its automatically generated by some pattern.
        :param external_key: Predefined external key that was already generated for another cloud account.
        May be used only for CI, tests and developers environments. Should not be used in production
        """
        method_url = self._api_root.settings.aws_cloud_accounts()
        request = {"arn_role": arn_role, "external_id": external_id}

        if arn_role is not None:
            print("arn_role " + arn_role)

        if name:
            request["name"] = name
        if external_key:
            request["id"] = external_key

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_azure_aks(self, name: str, cloud_account_name: str, kube_config: str):
        """ Register AKS"""
        method_url = self._api_root.settings.add_compute_service_aks()
        request = {
            "kube_config": kube_config,
            "name": name,
            "cloud_account_name": cloud_account_name,
        }

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_aws_k8s_unmanaged(self, name: str, cloud_account_name: str, kube_config: str):
        method_url = self._api_root.settings.add_compute_service_aws_k8s_unmanaged()
        request = {
            "kube_config": kube_config,
            "name": name,
            "cloud_account_name": cloud_account_name,
        }

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def update_aws_account(self, arn_role: str, external_id: str, name: str):
        """ Register AWS
        :param arn_role: IAM role ARN required for assume role to the customer cloud account
        :param external_id: External ID required for assume role to the customer cloud account
        :param name: Cloud account given name. If not passed its automatically generated by some pattern.
        May be used only for CI, tests and developers environments. Should not be used in production"""

        method_url = self._api_root.settings.aws_cloud_accounts()
        request = {"name": name, "arn_role": arn_role, "external_id": external_id}

        resp = self.session.put(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        return resp

    def get_account_template_url(self, external_id: str):
        """ Register AWS
        :param external_id: External ID required for assume role to the customer cloud account"""

        method_url = "{}?external_id={}".format(
            self._api_root.settings.aws_template(), external_id
        )
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        template_json = json.loads(resp.text)
        return template_json

    def unregister_cloud_account(self, cloud_account_name: str):
        method_url = self._api_root.settings.cloud_accounts()
        body = {"cloud_account_name": cloud_account_name}
        resp = self.session.delete(url=method_url, json=body)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_aws_s3_artifacts_repo(self, account_name: str, bucket_name: str):
        method_url = self._api_root.artifacts_repository.aws_s3_storage(
            space_name=self.space
        )
        request = {"account_name": account_name, "bucket_name": bucket_name}
        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_azure_artifact_repository(self, account_name: str, storage_name: str):
        method_url = self._api_root.artifacts_repository.azure_storage(
            space_name=self.space
        )
        request = {"account_name": account_name, "storage_name": storage_name}
        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_artifact_repository(self):
        method_url = self._api_root.artifacts_repository.artifact_repos(
            space_name=self.space
        )
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return Converters.create_artifact_repo_response(json.loads(res.text))

    def get_artifact_repo_storage(self, cloud_account: str):
        method_url = self._api_root.artifacts_repository.artifact_repos_by_cloud_account_name(
            cloud_account_name=cloud_account, space_name=self.space
        )
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return json.loads(res.text)

    def remove_artifact_repository(self):
        method_url = self._api_root.artifacts_repository.artifact_repos(
            space_name=self.space
        )
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def create_long_token(self) -> TokenResponse:
        method_url = self._api_root.token.long_token()
        resp = self.session.post(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        token_json = json.loads(resp.text)
        return Converters.create_token_response(token_json)

    def revoke_long_token(self, long_token: str):
        method_url = self._api_root.token.revoke_long_token()
        self.session.add_header("Authorization", "Bearer {}".format(long_token))
        resp = self.session.post(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def list_blueprints(
            self, exclude_samples=False, deployment_type: str = None
    ) -> List[BlueprintResponse]:
        params = {}
        if deployment_type:
            params["type"] = deployment_type
        method_url = self._api_root.blueprint.blueprints_by_space_name(self.space)
        resp = self.session.get(url=method_url, params=params)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        blueprints_json = json.loads(resp.text)
        all_blueprints = [
            Converters.create_blueprint_response(blueprint)
            for blueprint in blueprints_json
        ]
        if exclude_samples:
            all_blueprints = [
                blueprint for blueprint in all_blueprints if not blueprint.is_sample
            ]
        return all_blueprints

    def list_space_users(self) -> List[UserPermittedToSpaceResponse]:
        method_url = self._api_root.spaces.space_users(self.space)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        users_json = json.loads(resp.text)
        return [
            Converters.create_user_permitted_to_space_response(user)
            for user in users_json
        ]

    def publish_blueprint_in_catalog(self, blueprint_name: str):
        method_url = self._api_root.catalog.catalog(space_name=self.space)
        request = {"blueprint_name": blueprint_name}
        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def list_blueprint_files(
            self, blueprint_name: str, branch: str
    ) -> List[BlueprintFileResponse]:
        method_url = self._api_root.blueprint.blueprint_files(
            space_name=self.space, blueprint_name=blueprint_name, branch=branch
        )
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        files = json.loads(resp.text)
        return [Converters.create_blueprint_file_response(file) for file in files]

    def list_blueprints_in_catalog(self) -> List[CatalogForGetAllResponse]:
        method_url = self._api_root.catalog.catalog(space_name=self.space)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        blueprints_json = json.loads(resp.text)
        return [
            Converters.create_catalog_blueprint_response(blueprint)
            for blueprint in blueprints_json
        ]

    def unpublish_blueprint_in_catalog(self, blueprint_name: str):
        method_url = self._api_root.catalog.blueprint_in_catalog(
            space_name=self.space, blueprint_name=blueprint_name
        )
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_blueprint_details_in_catalog(
            self, blueprint_name: str
    ) -> CatalogForGetResponse:
        method_url = self._api_root.catalog.blueprint_in_catalog(
            space_name=self.space, blueprint_name=blueprint_name
        )
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        blueprint = json.loads(resp.text)
        return Converters.create_catalog_blueprint_full_response(blueprint)

    def list_sandboxes(self, retry: bool = True, **kwargs) -> List[SandboxResponseLean]:
        """
       :param retry: bool
       :param kwargs:  filter: str(automation/all/my),
                       sandbox_name: str,
                       count: int
       """

        method_url = self._api_root.sandbox.sandboxes()
        resp = self.session.get(url=method_url, params=kwargs)
        if resp.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.list_sandboxes(False, **kwargs)
            raise TimeoutError()
        GatewayUtils.handle_response(resp, return_codes=[200])
        result = json.loads(resp.text)
        return [
            Converters.create_sandbox_response_list_item(sandbox) for sandbox in result
        ]

    def list_productions(self, retry: bool = True) -> List[ProductionResponseLean]:
        url = self._api_root.production.productions()
        response = self.session.get(url=url)
        if response.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.list_productions(False)
            raise TimeoutError()
        GatewayUtils.handle_response(response, return_codes=[200])
        json_result = json.loads(response.text)
        return [
            Converters.create_production_response_list_item(item)
            for item in json_result
        ]

    def start_production(self, blueprint_name: str, artifacts: dict = None, inputs=None, compute_availability=None):
        url = self._api_root.production.productions()
        response = self.session.post(
            url=url,
            json={
                "blueprint_name": blueprint_name,
                "artifacts": artifacts if artifacts else {},
                "inputs": inputs if inputs else {},
                "compute_availability": compute_availability,
            },
        )
        GatewayUtils.handle_response(response=response, return_codes=[202])
        json_result = json.loads(response.text)
        return CreateProductionResponse(production_id=json_result["id"])

    def start_production_green(self,
                               production_id: str,
                               artifacts: dict = None,
                               inputs=None,
                               compute_availability: str = None):
        url = self._api_root.production.green_by_id(production_id=production_id)
        response = self.session.post(url=url, json={
            "artifacts": artifacts if artifacts else {},
            "inputs": inputs if inputs else {},
            "compute_availability": compute_availability
        })
        GatewayUtils.handle_response(response=response, return_codes=[202])
        json_result = json.loads(response.text)
        return CreateProductionResponse(production_id=json_result["id"])

    def update_production_debugging_service(self, production_id: str,
                                            environment: ProductionEnvironment,
                                            value: DebuggingServiceValue):
        if environment == ProductionEnvironment.BLUE:
            url = f'{self._api_root.production.blue_debugging_service(production_id=production_id)}'
        elif environment == ProductionEnvironment.GREEN:
            url = f'{self._api_root.production.green_debugging_service(production_id=production_id)}'
        else:
            raise ValueError(f'Unknown environment "{environment}"')

        response = self.session.put(url=f'{url}?value={value.value}')
        GatewayUtils.handle_response(response=response, return_codes=[200, 202])

    def expose_green(self, production_id: str, exposure_value: int):
        url = self._api_root.production.expose_green(production_id=production_id, exposure_value=exposure_value)
        response = self.session.put(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200, 202])

    def promote_green(self, production_id: str):
        url = self._api_root.production.promote_green(production_id=production_id)
        response = self.session.put(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200, 202])

    def start_sandbox(
            self, sandbox_name: str, blueprint_name: str, **kwargs
    ) -> CreateSandboxResponse:
        """
        :param sandbox_name: str
        :param blueprint_name: str
        :param kwargs:  artifacts: dict,
                        inputs: dict,
                        scheduled_end_time: datetime,
                        automation: bool,
                        duration: timedelta,
                        compute_availability: str,
                        lazy_load_artifacts: bool,
                        lazy_load_artifacts_timeout: int
        """
        method_url = self._api_root.sandbox.sandboxes()
        scheduled_end_time = (
            datetime.isoformat(kwargs.get("scheduled_end_time"))
            if kwargs.get("scheduled_end_time")
            else None
        )
        duration = duration_isoformat(kwargs.get("duration")) if kwargs.get("duration") else None
        resp = self.session.post(
            url=method_url,
            json={
                "blueprint_name": blueprint_name,
                "sandbox_name": sandbox_name,
                "artifacts": kwargs.get("artifacts", {}),
                "inputs": kwargs.get("inputs", {}),
                "automation": kwargs.get("automation", False),
                "scheduled_end_time": scheduled_end_time,
                "duration": duration,
                "compute_availability": kwargs.get("compute_availability"),
                "lazy_load_artifacts": kwargs.get("lazy_load_artifacts", False),
                "lazy_load_artifacts_timeout": kwargs.get("lazy_load_artifacts_timeout", 0),
            },
        )
        GatewayUtils.handle_response(response=resp, return_codes=[202])
        result = json.loads(resp.text)
        return CreateSandboxResponse(sandbox_id=result["id"])

    def end_sandbox(
            self,
            sandbox_id: str,
            retry: bool = True
    ):
        method_url = self._api_root.sandbox.sandbox_by_id(sandbox_id=sandbox_id)
        resp = self.session.delete(url=method_url)

        if resp.status_code in [408, 429, 504]:
            if retry:
                time.sleep(30)  # need to make sure that the next call will not use the cached timedout error
                self.end_sandbox(sandbox_id, False)
                return

        GatewayUtils.handle_response(response=resp, return_codes=[202])

    def end_production(self, production_id: str):
        method_url = self._api_root.production.production_by_id(
            production_id=production_id
        )
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[202])

    def end_production_green(self, production_id: str):
        method_url = self._api_root.production.green_by_id(production_id=production_id)
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[202])

    def get_sandbox_details(
            self, sandbox_id: str, retry: bool = True
    ) -> Optional[SandboxResponse]:
        method_url = self._api_root.sandbox.sandbox_by_id(sandbox_id=sandbox_id)
        resp = self.session.get(url=method_url)
        if resp.status_code == 404:
            return None
        if resp.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.get_sandbox_details(sandbox_id, False)
            raise TimeoutError()

        GatewayUtils.handle_response(response=resp, return_codes=[200])
        result = json.loads(resp.text)
        sandbox_response = Converters.create_sandbox_response(result)
        return sandbox_response

    def get_production_details(
            self, production_id: str, retry: bool = True) \
            -> Optional[ProductionBlueResponse]:
        url = self._api_root.production.production_by_id(production_id)
        response = self.session.get(url=url)
        if response.status_code == 404:
            return None
        if response.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.get_production_details(production_id, False)
            raise TimeoutError()

        GatewayUtils.handle_response(response=response, return_codes=[200])
        json_result = json.loads(response.text)
        production_response = Converters.create_production_blue_response(json_result)
        return production_response

    def get_production_green_details(
            self,
            production_id: str,
            retry: bool = True
    ) -> Optional[ProductionGreenResponse]:
        url = self._api_root.production.green_by_id(production_id)
        response = self.session.get(url=url)
        if response.status_code == 404:
            return None
        if response.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.get_production_green_details(production_id, False)
            raise TimeoutError()

        GatewayUtils.handle_response(response=response, return_codes=[200])
        json_result = json.loads(response.text)
        production_response = Converters.create_production_green_response(json_result)
        return production_response

    def add_repository_to_space(
            self,
            repository_details: BlueprintRepositoryDetails,
            deployment_type: str = None,
            wait_for_cache_refresh: bool = False,
    ):
        method_url = self._api_root.spaces.space_repositories(space_name=self.space)
        request = {
            "repository_url": repository_details.repository_url,
            "access_token": repository_details.access_token,
            "repository_type": repository_details.repository_type,
            "branch": repository_details.branch,
            "type": deployment_type
        }
        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

        if wait_for_cache_refresh:
            Utils.wait_for(
                func=lambda: (not self.list_blueprints(exclude_samples=True) is False),
                interval_sec=5,
                max_retries=24,
                subject="blueprints ready",
            )

    def remove_repository_from_space(self, wait_for_cache_refresh=False):
        method_url = self._api_root.spaces.space_repositories(space_name=self.space)
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

        if wait_for_cache_refresh:
            Utils.wait_for(
                func=lambda: (not self.list_blueprints(exclude_samples=True) is True),
                interval_sec=5,
                max_retries=24,
                subject="blueprints removed",
            )

    def get_tfstate_file(self, sandbox_id: str, service_name: str) -> bytes:
        method_url = self._api_root.sandbox.sandbox_tfstate(sandbox_id=sandbox_id, service_name=service_name)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        return resp.content

    def validate_blueprint(self, blueprint: str, branch: str = None) -> bool:
        method_url = "{base}?blueprint={blueprint}&branch={branch}".format(
            base=self._api_root.blueprint_validation.blueprint_validation_by_space_name(
                space_name=self.space
            ),
            blueprint=blueprint,
            branch=branch if branch else "master",
        )
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        if resp.json() == {"errors": []}:
            return True
        return resp.json()

    def invite_user(self, invites: CreateInvitationsRequest):
        method_url = self._api_root.account.invitations()
        resp = self.session.post(url=method_url, json=invites.__dict__)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_invites(self, space_name: str = None) -> List[UserInvitationResponse]:
        method_url = self._api_root.account.account_invitations(space=space_name)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        result = json.loads(resp.text)
        return [Converters.create_user_invitation_response(x) for x in result]

    def test_invite_exists(self, secret: str):
        method_url = self._api_root.account.account_signup_by_secret(secret=secret)
        resp = self.session.get(url=method_url)
        return resp.status_code

    def remove_invites(self, emails: List[str]):
        method_url = self._api_root.account.account_invitations()
        req = {"emails": emails}
        resp = self.session.delete(url=method_url, json=req)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def list_repositories(
            self, repository_type: str = None
    ) -> List[RepositoryResponse]:
        params = {}
        if repository_type:
            params["type"] = repository_type
        url = self._api_root.spaces.space_repositories(space_name=self.space)
        response = self.session.get(url=url, params=params)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        json_result = json.loads(response.text)
        return [Converters.create_repository_response(x) for x in json_result]

    def update_scheduled_end_time(self, sandbox_id: str, scheduled_end_time: datetime):
        url = "{base}?value={end_time}".format(
            base=self._api_root.sandbox.sandbox_scheduled_end_time(
                sandbox_id=sandbox_id
            ),
            end_time=quote(scheduled_end_time.isoformat()),
        )
        response = self.session.put(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def turn_on_debugging_service(self, sandbox_id: str):
        self.update_debugging_service(sandbox_id=sandbox_id, value=DebuggingServiceValue.ON)

    def turn_off_debugging_service(self, sandbox_id: str):
        self.update_debugging_service(sandbox_id=sandbox_id, value=DebuggingServiceValue.OFF)

    def update_debugging_service(self, sandbox_id: str, value: DebuggingServiceValue):
        url = self._api_root.sandbox.sandbox_debugging_service(sandbox_id=sandbox_id)
        response = self.session.put(url=f'{url}?value={value.value}')
        GatewayUtils.handle_response(response=response, return_codes=[200, 202])

    def get_space_achievements(self) -> dict:
        url = self._api_root.achievements.achievements(space_name=self.space)
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response, [200])
        return json.loads(response.text)

    def create_space(self, space_name: str, cloud_accounts: []):
        url = self._api_root.spaces.spaces()
        request = {"name": space_name, "cloud_accounts": cloud_accounts}

        response = self.session.post(url=url, json=request)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def delete_space(self, space_name: str):
        url = self._api_root.spaces.space_by_name(space_name=space_name)
        response = self.session.delete(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_space(self, name: str) -> GetSpaceResponse:
        url = self._api_root.spaces.space_by_name(space_name=name)
        res = self.session.get(url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return GetSpaceResponse(json.loads(res.text))

    def get_images(self, space_name: str, cloud_account_name: str, region_id: str):
        url = self._api_root.spaces.get_images(space_name, cloud_account_name, region_id)
        res = self.session.get(url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return [Converters.get_image_response(x) for x in json.loads(res.text)]

    def update_space(self, space: str, update_space_request: UpdateSpaceRequest):
        url = self._api_root.spaces.space_by_name(space_name=space)
        response = self.session.put(url=url, json=update_space_request.__dict__)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_space_roles(self) -> List[RoleListItemResponse]:
        url = self._api_root.settings.space_roles()
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        roles_json = json.loads(response.text)
        return [Converters.create_role_list_item_response(role) for role in roles_json]

    def get_account_roles(self) -> List[RoleListItemResponse]:
        url = self._api_root.settings.account_roles()
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        roles_json = json.loads(response.text)
        return [Converters.create_role_list_item_response(role) for role in roles_json]

    def remove_cloud_account_from_space(self, cloud_account: str):
        method_url = self._api_root.spaces.space_cloud_account(
            self.space, cloud_account
        )
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def add_cloud_account_to_space(self, cloud_account: str,
                                   infra_settings: SpaceCloudAccountInfraSettingsRequest = None):
        method_url = self._api_root.spaces.space_cloud_accounts(self.space)
        data = {"name": cloud_account}
        if infra_settings:
            data['internet_facing'] = infra_settings.internet_facing
            data['existing_infra'] = infra_settings.existing_infra
            if infra_settings.existing_infra_sets:
                existing_infra_sets = []
                for infra_set in infra_settings.existing_infra_sets:
                    existing_infra_sets.append(
                        {
                            'region': infra_set.region,
                            'virtual_network': infra_set.virtual_network,
                            'subnets': {'gateway_subnet': infra_set.subnets.gateway_subnet,
                                        'management_subnet': infra_set.subnets.management_subnet,
                                        'application_subnets': infra_set.subnets.application_subnets
                                        } if infra_set.subnets else None
                        }
                    )
                data['existing_infra_sets'] = existing_infra_sets
        res = self.session.post(url=method_url, json=data)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def update_cloud_account_in_space(self, cloud_account: str, infra_settings: SpaceCloudAccountInfraSettingsRequest):
        method_url = self._api_root.spaces.space_cloud_account(space_name=self.space, cloud_account=cloud_account)
        data = {
            'internet_facing': infra_settings.internet_facing,
            'existing_infra': infra_settings.existing_infra
        }
        if infra_settings.existing_infra_sets:
            existing_infra_sets = []
            for infra_set in infra_settings.existing_infra_sets:
                existing_infra_sets.append(
                    {
                        'region': infra_set.region,
                        'virtual_network': infra_set.virtual_network,
                        'subnets': {'gateway_subnet': infra_set.subnets.gateway_subnet,
                                    'management_subnet': infra_set.subnets.management_subnet,
                                    'application_subnets': infra_set.subnets.application_subnets
                                    } if infra_set.subnets else None
                    }
                )
            data['existing_infra_sets'] = existing_infra_sets
        res = self.session.put(url=method_url, json=data)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def get_cloud_accounts_under_space(self) -> List[CloudAccountInSpaceResponse]:
        method_url = self._api_root.spaces.space_cloud_accounts(self.space)
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return [
            Converters.create_cloud_account_in_space_response(x) for x in json.loads(res.text)
        ]

    def get_account_users(self) -> List[UserForAllUsersResponse]:
        method_url = self._api_root.account.account_users()
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        users_json = json.loads(res.text)
        return [Converters.create_user_for_all_users_response(x) for x in users_json]

    def get_space_permissions(self) -> List[str]:
        method_url = self._api_root.spaces.user_permissions(self.space)
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return json.loads(res.text)

    def delete_account(self, account_name: str):
        method_url = self._api_root.account.delete_account(account_name)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return res.text

    def contact_me(self, first_name: str, last_name: str, email: str, phone: str):
        method_url = self._api_root.account.contact_me()
        request = {
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
            "phone": phone
        }
        res = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def add_user_to_space(self, user_email: str, space_role: str):
        method_url = self._api_root.spaces.space_users(self.space)
        request = {"email": user_email, "space_role": space_role}
        res = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def update_user_space_role(self, user_email: str, space_role: str):
        method_url = self._api_root.spaces.user_space_role(self.space, user_email, space_role)
        res = self.session.put(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def remove_user_from_space(self, user_email: str):
        method_url = self._api_root.spaces.space_user(self.space, user_email)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def update_user_account_role(self, user_email: str, account_role: str):
        method_url = self._api_root.account.account_user_account_role_value(user_email, account_role)
        res = self.session.put(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def remove_user_account_role(self, user_email: str):
        method_url = self._api_root.account.account_user_account_role(user_email)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def get_access_url_by_link(self, access_link: AccessLink) -> str:
        method_url = f"{self.api_address}{access_link.link}"
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        res_json = json.loads(res.text)
        return res_json['url']

    def get_access_url_by_params(self, sandbox_id: str, instance_id: str, protocol: str) -> str:
        method_url = self._api_root.qualiy.connect(sandbox_id=sandbox_id, instance_id=instance_id, protocol=protocol)
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        res_json = json.loads(res.text)
        return res_json['url']

    def delete_user(self, user_email: str):
        method_url = self._api_root.account.account_user(user_email)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def add_artifactory_repo(self, server_url: str, artifactory_username: str, api_key: str):
        method_url = self._api_root.artifacts_repository.artifactory(space_name=self.space)
        request = {
            "server_url": server_url,
            "artifactory_username": artifactory_username,
            "api_key": api_key
        }
        response = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_cost_breakdown(self, frm: datetime, until: datetime, criteria: str = 'blueprint') -> dict:
        url = self._api_root.cost.breakdown(space_name=self.space, frm=frm, until=until, criteria=criteria)
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response, [200])
        return json.loads(response.text)

    def get_cost_usage(self, date: datetime, timezone: str = '') -> dict:
        url = self._api_root.cost.usage(space_name=self.space, date=date, timezone=timezone)
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response, [200])
        return json.loads(response.text)

    def post_request(self, url: str, request_json: dict, return_codes: List = None):
        return_codes = return_codes or [200]
        response = self.session.post(url=url, json=request_json)
        GatewayUtils.handle_response(response=response, return_codes=return_codes)

    def put_request(self, url: str, request_json: dict, return_codes: List = None):
        return_codes = return_codes or [200]
        response = self.session.put(url=url, json=request_json)
        GatewayUtils.handle_response(response=response, return_codes=return_codes)

    def set_account_extra_details(self, account_name: str, sso_enabled: bool, anonymous_users: bool):
        set_account_extra_details_url = self._api_root.account.set_account_extra_details(account_name)
        self.put_request(set_account_extra_details_url, {"sso_enabled": sso_enabled, "anonymous_users": anonymous_users}
                         , return_codes=[200])

    def get_account_extra_details(self, account_name: str):
        set_account_extra_details_url = self._api_root.account.set_account_extra_details(account_name)
        response = self.session.get(url=set_account_extra_details_url)
        GatewayUtils.handle_response(response, [200])
        return json.loads(response.text)

    def get_space_cloud_account_regions(self, space_name: str, cloud_account_name: str) -> List[RegionResponse]:
        method_url = self._api_root.spaces.space_cloud_account_regions(space_name=space_name,
                                                                       cloud_account_name=cloud_account_name)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        regions_json = json.loads(resp.text)
        return [Converters.create_space_cloud_account_regions_response(region) for region in regions_json]

    def verify_cloud_account(self, cloud_account_name: str) -> VerifyCloudProviderResponse:
        method_url = self._api_root.settings.verify_cloud_account()
        data = {"cloud_account_name": cloud_account_name}
        res = self.session.put(url=method_url, json=data)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        verify_json = json.loads(res.text)
        return Converters.create_verify_response(verify_json)


class SandboxesHandler:
    def __init__(self):
        self.sandboxes_set = set()

    def add_sandbox(self, sandbox_id):
        self.sandboxes_set.add(sandbox_id)
        print("Sandbox {id} added to cleaner".format(id=sandbox_id))

    def remove_sandbox(self, sandbox_id):
        self.sandboxes_set.remove(sandbox_id)
        print("Sandbox {id} removed from cleaner".format(id=sandbox_id))


class SandboxCleaner:
    def __init__(self, client: Colony, end_on_exception=True):
        self.end_on_exception = end_on_exception
        self.client = client
        self.sandboxes = SandboxesHandler()
        self.space = self.client.space

    def __enter__(self):
        return self.sandboxes

    def __exit__(self, exc_type, exc_val: str, exc_tb):
        for sandbox in self.sandboxes.sandboxes_set:
            # if exc_type is MaxRetriesException:
            #     print(f"Sandbox {sandbox} failed on timeout")
            if exc_type is SandboxNotFound:
                print(f"Sandbox {sandbox} failed on not found")
            elif exc_type is SandboxEndingFailed:
                print(f"Sandbox {sandbox} failed to end")
            else:
                if self.end_on_exception:
                    print("Sandbox {} is being removed by cleaner".format(sandbox))
                    self.client.end_sandbox(sandbox_id=sandbox)


class ProductionCleaner:
    def __init__(self, client: Colony):
        self.client = client
        self.ids = []
        self.space = self.client.space

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for production_id in self.ids:
            print("Production {} is being removed by cleaner".format(production_id))

            self.client.end_production(production_id=production_id)

    def add(self, production_id):
        self.ids.append(production_id)
        print("Production {id} added to cleaner".format(id=production_id))

    def remove(self, production_id):
        self.ids.remove(production_id)
        print("Production {id} removed from cleaner".format(id=production_id))



"""
Container for Error classes used in cs18-api-client
"""
import json

from requests import Response
from typing import List


class Error:
    def __init__(self, json_data: dict):
        self.code = json_data["code"]
        self.name = json_data["name"]
        self.message = json_data["message"]

    def __eq__(self, other):
        if isinstance(other, Error):
            return (
                self.code == other.code
                and self.name == other.name
                and self.message == other.message
            )
        return False


class SandboxError:
    def __init__(self, time: str, code: str, message: str):
        self.message = message
        self.code = code
        self.time = time


class GeneralHttpException(Exception):
    """
    General HTTP Exception.
    Loads additional information we want to be passed with the error.
    """

    def __init__(self, response: Response):
        super(GeneralHttpException, self).__init__(
            "Request to '{url}' ended with wrong status_code '{status_code}'. "
            "Message {message}. "
            "TraceID: {trace}.".format(
                url=response.url,
                status_code=response.status_code,
                message=response.text,
                trace=response.headers.get("X-Correlation-ID"),
            )
        )
        self.status_code = response.status_code
        self.trace_id = response.headers.get("X-Correlation-ID")


class DevboxHttpException(GeneralHttpException):
    """
    Devbox HTTP Exception
    """

    def __init__(self, response: Response):
        super(DevboxHttpException, self).__init__(response)
        self.errors = [Error(x) for x in json.loads(response.text).get("errors")]


class UnauthorizedException(Exception):
    def __init__(
            self,
            credentials: dict
    ):
        super(UnauthorizedException, self).__init__(
            "Failed to authenticate Colony client. Should provide either an access_token or both email and password."
            "Provided credentials: {credentials}".format(credentials=credentials)
        )


class MaxRetriesException(Exception):
    def __init__(self, max_retries, subject):
        super(MaxRetriesException, self).__init__(
            f"Timeout: {max_retries} attempts were exhausted waiting for '{subject}'"
        )


class SandboxEndingFailed(Exception):
    def __init__(self, sandbox_id: str, ending_errors: List[SandboxError]):
        super(SandboxEndingFailed, self).__init__(
            "Sandbox '{SANDBOX_ID}' failed to end with those errors: {ERRORS}".format(
                SANDBOX_ID=sandbox_id,
                ERRORS=", ".join("{CODE} ({MESSAGE})".format(CODE=error.code,
                                                             MESSAGE=error.message)
                                 for error in ending_errors))
        )


class SandboxNotFound(Exception):
    pass



"""
Container for Request object used in cs18-api-client
"""
from enum import Enum
from typing import List


class ProductionEnvironment(Enum):
    BLUE = 'blue'
    GREEN = 'green'


class DebuggingServiceValue(Enum):
    ON = 'on'
    OFF = 'off'


class SpaceCloudAccountSubnetsRequest:
    def __init__(self, gateway_subnet: str = None,
                 management_subnet: str = None,
                 application_subnets: List[str] = None):
        self.gateway_subnet = gateway_subnet
        self.management_subnet = management_subnet
        self.application_subnets = application_subnets


class SpaceCloudAccountInfraSetRequest:
    def __init__(self,
                 region: str,
                 virtual_network: str,
                 subnets: SpaceCloudAccountSubnetsRequest):
        self.region = region
        self.virtual_network = virtual_network
        self.subnets = subnets


class SpaceCloudAccountInfraSettingsRequest:
    def __init__(self,
                 internet_facing: bool,
                 existing_infra: bool,
                 existing_infra_sets: List[SpaceCloudAccountInfraSetRequest] = None):
        self.internet_facing = internet_facing
        self.existing_infra = existing_infra
        self.existing_infra_sets = existing_infra_sets


class UserSignupRequest:
    def __init__(self, first_name: str, last_name: str, password: str, secret: str):
        self.first_name = first_name
        self.last_name = last_name
        self.password = password
        self.secret = secret


class CreateInvitationsRequest:
    def __init__(
            self,
            emails: [],
            account_role: str,
            reason: str,
            space_name: str,
            space_role: str,
    ):
        self.emails = emails
        self.account_role = account_role
        self.reason = reason
        self.space_name = space_name
        self.space_role = space_role


class CreateAccountRequest:
    def __init__(
            self,
            account_name: str,
            first_name: str,
            last_name: str,
            email: str,
            password: str,
            phone_number: str,
    ):
        self.phone_number = phone_number
        self.password = password
        self.email = email
        self.last_name = last_name
        self.first_name = first_name
        self.account_name = account_name


class UpdateSpaceRequest:
    def __init__(self, name: str):
        self.name = name



"""
Container for object classes used in cs18-api-client
"""
from datetime import datetime
from typing import Optional

import dateutil.parser

from gateways.email_utils import EmailAccount


class AccessLink:
    def __init__(self, protocol: str, link: str):
        self.link = link
        self.protocol = protocol


class CloudAccountCostResponse:
    def __init__(self, last_update: Optional[datetime]):
        self.last_update = last_update


class CloudComputeService:
    def __init__(
            self,
            name: str,
            service_type: str,
            created_date: datetime,
            created_by: str,
            user_defined: bool,
    ):
        self.name = name
        self.type = service_type
        self.created_date = created_date
        self.created_by = created_by
        self.user_defined = user_defined


class Commit:
    def __init__(self, data: dict):
        self.sha = data["sha"]
        self.author = data["commit"]["author"]["name"]
        self.date = dateutil.parser.parse(data["commit"]["author"]["date"])
        self.comment = data["commit"]["message"]

    def __str__(self):
        return "{0}: [{1}] {2}".format(self.date, self.sha[:7], self.comment)


class ColonyAccount:
    def __init__(
            self, account: str, email: str, password: str, first_name: str, last_name: str,
            email_account: EmailAccount = None
    ):
        self.account = account
        self.default_space = "Trial"
        self.email = email
        self.password = password
        self.first_name = first_name
        self.last_name = last_name
        self.email_account = email_account


class BlueprintRepositoryDetails:
    def __init__(self, repository_url: str, access_token: str, repository_type: str, branch: str = None):
        self.repository_url = repository_url
        self.repository_type = repository_type
        self.access_token = access_token
        self.branch = branch



from gateways.apis.api_base_class import ApiBase


class ApiBlueprintValidation(ApiBase):
    def blueprint_validation_by_space_name(self, space_name: str):
        return self.build_route(
            "spaces/{space_name}/validation/blueprint".format(**locals())
        )



from gateways.apis.api_base_class import ApiBase


class ApiSettings(ApiBase):
    def cloud_providers(self):
        return self.build_route("settings/cloudproviders")

    def cloud_accounts(self):
        return self.build_route("settings/cloudaccounts")

    def compute_services(self):
        return self.build_route("settings/cloudaccounts/computeservices")

    def status(self):
        return self.build_route("settings/status")

    def azure_cloud_accounts(self):
        return self.build_route("settings/cloudaccounts/azure")

    def aws_cloud_accounts(self):
        return self.build_route("settings/cloudaccounts/aws")

    def add_compute_service_aks(self):
        return self.build_route("settings/cloudaccounts/azure/computeservices/aks")

    def add_compute_service_aws_k8s_unmanaged(self):
        return self.build_route("settings/cloudaccounts/aws/computeservices/k8s_unmanaged")

    def aws_template(self):
        return self.build_route("settings/cloudaccounts/aws/template")

    def space_roles(self):
        return self.build_route("settings/spaceroles")

    def account_roles(self):
        return self.build_route("settings/accountroles")

    def verify_cloud_account(self):
        return self.build_route("settings/cloudaccounts/verify")



import datetime
import json
import os
from typing import List

import time
from requests import Response
from gateways.common.cs18_api_errors import DevboxHttpException, GeneralHttpException, MaxRetriesException
from gateways.common.cs18_api_responses import (
    ApplicationResponse,
    CatalogApplicationResponse,
    ProductionResponse,
    SandboxResponse,
    ServiceResponse
)
from gateways.common.cs18_api_statuses import SandboxStatus, ServiceStatus


class SandboxUtils:
    """
    Methods for sandbox information retrieval
    """

    @staticmethod
    def get_app_artifacts_in_common_s3(
            apps: List[CatalogApplicationResponse],
            version: str = "production"
    ) -> dict:
        """
        Generates full path to artifact filename in S3 for given List[CatalogApplicationResponse]
        :param apps: List[CatalogApplicationResponse]
        :param version: str. Default: "production"
        :return: dict
        """
        artifacts = {}
        for app in apps:
            if "windows" in app.name:
                artifacts.update(
                    {
                        app.name: "{app_name}/{version}/{app_name}.zip".format(
                            app_name=app.name, version=version
                        )
                    }
                )
            else:
                artifacts.update(
                    {
                        app.name: "{app_name}/{version}/{app_name}.tar.gz".format(
                            app_name=app.name, version=version
                        )
                    }
                )

        return artifacts

    @staticmethod
    def get_app_by_name(
            applications: List[ApplicationResponse], service_name: str
    ) -> ApplicationResponse:
        """
        Returns a specific ApplicationResponse from a list by given name
        :param applications: List[ApplicationResponse]
        :param service_name: specific service name to look for
        :return: ApplicationResponse
        """
        return next(
            iter(
                [
                    application
                    for application in applications
                    if application.name == service_name
                ]
            ),
            None,
        )

    @staticmethod
    def is_deployment_ended(status: str) -> bool:
        """
        True if sandbox deployment has ended ActiveWithError or Active
        :param status: str
        :return: bool
        """
        return SandboxUtils.is_deployment_succeeded(status=status) or SandboxUtils.is_deployment_failed(status=status)

    @staticmethod
    def is_deployment_succeeded(status: str) -> bool:
        """
        True if sandbox status is Active
        :param status: str
        :return: bool
        """
        return status == SandboxStatus.ACTIVE

    @staticmethod
    def is_deployment_failed(status: str) -> bool:
        """
        True if sandbox status is ActiveWithError
        :param status: str
        :return: bool
        """
        return status == SandboxStatus.ACTIVE_WITH_ERROR

    @staticmethod
    def describe_sandbox_status(sandbox: SandboxResponse) -> str:
        """
        Generates string response of sandbox status
        :param sandbox: SandboxResponse
        :return: str
        """
        return "sandbox status: {STATUS} ({STATUS_DETAILS}), " \
               "app statuses: {APP_STATUSES}, " \
               "service statuses: {SERVICE_STATUSES}" \
               "sandbox errors: [{ERRORS}]".format(STATUS=sandbox.sandbox_status,
                                                   STATUS_DETAILS=sandbox.status_details,
                                                   APP_STATUSES=", ".join(
                                                       "{APP_NAME}: {APP_STATUS}".format(
                                                           APP_NAME=app.name, APP_STATUS=app.status
                                                       )
                                                       for app in sandbox.applications
                                                   ),
                                                   SERVICE_STATUSES=", ".join(
                                                       f"{service.name}: {service.status}"
                                                       for service in sandbox.services
                                                   ),
                                                   ERRORS=", ".join(
                                                       "{CODE} ({MESSAGE})".format(CODE=error.code,
                                                                                   MESSAGE=error.message)
                                                       for error in sandbox.errors
                                                   ),
                                                   )

    @staticmethod
    def describe_production_status(production: ProductionResponse) -> str:
        return (
            "Production status: {STATUS} ({STATUS_DETAILS}), update_status: {UPDATE_STATUS}"
            "app statuses: {APP_STATUSES}, sandbox errors: [{ERRORS}]".format(
                STATUS=production.production_status,
                STATUS_DETAILS=production.status_details,
                UPDATE_STATUS=production.update_status,
                APP_STATUSES=", ".join(
                    "{APP_NAME}: {APP_STATUS}".format(
                        APP_NAME=app.name, APP_STATUS=app.status
                    )
                    for app in production.applications
                ),
                ERRORS=", ".join(
                    "{CODE} ({MESSAGE})".format(CODE=error.code, MESSAGE=error.message)
                    for error in production.errors
                ),
            )
        )

    @staticmethod
    def is_service_deployment_ended(service: ServiceResponse):
        return service.status not in [ServiceStatus.PENDING, ServiceStatus.SETUP]


class ErrorUtils:
    @staticmethod
    def get_exception_error_codes(exception: DevboxHttpException) -> List[str]:
        return [err.code for err in exception.errors]

    @staticmethod
    def format_errors(exception: DevboxHttpException) -> str:
        return ",".join(f"{err.code} [{err.message}]" for err in exception.errors)


class Utils:
    """
    Various helper methods
    """

    @staticmethod
    def wait_for(
            func,
            interval_sec: int = 10,
            max_retries: int = 10,
            subject: str = "",
            silent: bool = False
    ):
        """
        Runs a given function in a loop until max_retries have been exhausted
        :param func: function to perform in a loop
        :param interval_sec: interval in seconds
        :param max_retries: max retries before exiting the loop
        :param subject: action to be performed, for human readability
        :param silent: print to stdout, True/False
        :return: None or Exception
        """
        current_attempt = 1
        while True:
            result = func()
            predicate_success, more_info = (
                result if isinstance(result, tuple) else (result, None)
            )
            if not silent:
                print(
                    "[{TIMESTAMP} utc] Attempt {ATTEMPT}/{MAX_ATTEMPTS} {CHECK_RESULT}{SUBJECT}{MORE_INFO}".format(
                        TIMESTAMP=Utils.get_current_time_utc().strftime("%X"),
                        ATTEMPT=current_attempt,
                        MAX_ATTEMPTS=max_retries,
                        CHECK_RESULT="succeeded" if predicate_success else "failed",
                        SUBJECT=" for '{}'".format(subject) if subject else "",
                        MORE_INFO=", {}".format(more_info) if more_info else "",
                    )
                )
            if predicate_success:
                return

            current_attempt += 1
            if current_attempt > max_retries:
                raise MaxRetriesException(max_retries, subject)

            time.sleep(interval_sec)

    @staticmethod
    def wait_until_time(
            func,
            interval_sec: int = 10,
            max_seconds: int = 300,  # 5 min
            subject: str = "",
            silent: bool = False
    ):
        """
        Runs a given function in a loop until max_retries have been exhausted
        :param func: function to perform in a loop
        :param interval_sec: interval in seconds
        :param max_seconds: max seconds before exiting the loop
        :param subject: action to be performed, for human readability
        :param silent: print to stdout, True/False
        :return: None or Exception
        """
        current_time = time.time()
        timeout = current_time + max_seconds

        while current_time < timeout:
            result = func()
            predicate_success, more_info = (
                result if isinstance(result, tuple) else (result, None)
            )
            if not silent:
                print(
                    "[{TIMESTAMP} utc] Attempt {ATTEMPT}/{MAX_ATTEMPTS_SECONDS} "
                    "{CHECK_RESULT}{SUBJECT}{MORE_INFO}".format(
                        TIMESTAMP=Utils.get_current_time_utc().strftime("%X"),
                        ATTEMPT=(max_seconds - round(timeout - current_time)),
                        MAX_ATTEMPTS_SECONDS=max_seconds,
                        CHECK_RESULT="succeeded" if predicate_success else "failed",
                        SUBJECT=" for '{}'".format(subject) if subject else "",
                        MORE_INFO=", {}".format(more_info) if more_info else "",
                    )
                )
            if predicate_success:
                return

            current_time = time.time()
            time.sleep(interval_sec)

        if current_time > timeout:
            raise MaxRetriesException(max_seconds, subject)

    @staticmethod
    def get_current_time_utc() -> datetime:
        """
        Returns current datetime in UTC
        :return: datetime
        """
        return datetime.datetime.now(datetime.timezone.utc)

    @staticmethod
    def find_dir(file: str) -> str:
        """
        Helper method to find specific file under current working directory
        :param file: str
        :return: str
        """
        for dirpath, _, filenames in os.walk(os.getcwd()):
            if file in filenames:
                return dirpath
        return "{} was not found".format(file)


class GatewayUtils:
    """
    Various helper methods.
    Possibly move handle_response to more appropriate place,
        since it's an method for handling exception generation
    """

    @staticmethod
    def get_cs18_api_address(provider: str, host: str = None) -> str:
        """
        Returns CS18 API endpoint, either from config_file,
        or host if provided. Default: http://localhost:5050
        :param provider: str
        :param host: optional str
        :return: str
        """
        if host:
            return host

        data = "test_execution_data_{}".format(provider)
        config_file = "{dir}/{file}".format(dir=Utils.find_dir(data), file=data)
        if os.path.isfile(config_file):
            with open(config_file, "r") as file:
                config_json = json.load(file)
                for service in config_json["applications"]:
                    if service["name"] == "cs18-api":
                        return service.get("shortcuts")[0]
        return "http://localhost:5050"

    @staticmethod
    def get_cs18_web_address(provider: str, host: str = None) -> str:
        """
        Returns CS18 web address, either from config_file,
        or host if provided. Default: http://localhost
        :param provider: str
        :param host: optional str
        :return: str
        """
        if host:
            return host

        data = "test_execution_data_{}".format(provider)
        config_file = "{dir}/{file}".format(dir=Utils.find_dir(data), file=data)
        if os.path.isfile(config_file):
            with open(config_file, "r") as file:
                config_json = json.load(file)
                for service in config_json["applications"]:
                    if service["name"] == "cs18-ui":
                        return service.get("shortcuts")[0]
        return "http://localhost"

    @staticmethod
    def handle_response(response: Response, return_codes: List[int]):
        """
        Raises appropriate exception when response's
        return code is not found in return_codes list
        :param response: Response
        :param return_codes: List[int]
        :return: Exception if needed
        """
        if response.status_code not in return_codes:
            try:
                raise DevboxHttpException(response)
            except (json.JSONDecodeError, AttributeError):
                raise GeneralHttpException(response)



class ApiBase:
    def __init__(self, api_address: str, version: str = None):
        _version = "/v{}".format(version) if version else ""
        self._root = "{api_address}/api{_version}".format(**locals())

    def build_route(self, route: str):
        return "{root}/{route}".format(root=self._root, route=route)



import os
import shutil


class ArtifactPublisher:
    ARTIFACTS_FOLDER_PATH = "/tmp/artifact"
    current_test_path = ""

    @staticmethod
    def init_artifact_folder(test_method_name):
        ArtifactPublisher._create_artifact_folder(test_method_name)

    @staticmethod
    def _delete_artifact_folder(only_current_test: bool = False):
        folder_path = ArtifactPublisher.ARTIFACTS_FOLDER_PATH
        if only_current_test:
            folder_path = ArtifactPublisher.current_test_path
        if os.path.exists(folder_path):
            shutil.rmtree("/{}".format(folder_path), ignore_errors=True)
            print("Deleted old artifacts folder: '{}'".format(folder_path))

    @staticmethod
    def _create_artifact_folder(test_method_name):
        folder_path = ArtifactPublisher.ARTIFACTS_FOLDER_PATH
        ArtifactPublisher.current_test_path = os.path.join(folder_path, test_method_name)
        os.makedirs(ArtifactPublisher.current_test_path)
        print(
            "Created artifacts folder: '{}'".format(ArtifactPublisher.current_test_path)
        )

    @staticmethod
    def get_next_free_filename(file_name):
        folder_path = ArtifactPublisher.current_test_path
        file_ext = os.path.splitext(file_name)[1]
        file_name_no_ext = os.path.splitext(file_name)[0]

        if not os.path.exists(os.path.join(folder_path, file_name)):
            return file_name
        i = 1
        while True:
            new_file_name = "{}({}){}".format(file_name_no_ext, i, file_ext)
            if os.path.exists(os.path.join(ArtifactPublisher.current_test_path, new_file_name)):
                i += 1
            else:
                break
        return new_file_name



import re
import time
from random import randrange
from typing import List, Callable
import os
from guerrillamail import GuerrillaMailSession

from gateways.email_utils.artifact_publisher import ArtifactPublisher
from gateways.email_utils.nada import Nada


#
# This implementation wrap the GuerrillaMail and GetNada functionality
# We may change it when staring to use a different mail service
#


class MailSession:
    def __init__(self, session):
        self.session = session

    def get_email_list(self):
        return self.session.get_email_list()

    def get_email(self, _id):
        return self.session.get_email(id)


class EmailMessage:
    def __init__(self, session: MailSession, email_id, subject, sender, datetime):
        self._id = email_id
        self.subject = subject
        self.sender = sender
        self.datetime = datetime
        self.body = None
        self._body_fetched = False
        self._session = session

    def __getattribute__(self, item):
        if item == "body":
            if not self._body_fetched:
                email = self._session.get_email(self._id)
                self.body = email.body
                self._body_fetched = True
        return super(EmailMessage, self).__getattribute__(item)


class EmailAccount:
    def __init__(self, session, email_address: str):
        self.email_address = email_address
        self._session = session

    def get_emails(self) -> List[EmailMessage]:
        all_items = []
        more = True
        while more:
            bulk_items = self._session.get_email_list()
            all_items.extend(bulk_items)
            more = len(bulk_items) >= 20

        return [
            EmailMessage(self._session, x.guid, x.subject, x.sender, x.datetime)
            for x in all_items
        ]


class EmailAccountFactory:
    @staticmethod
    def create_random(mail_type: str) -> EmailAccount:
        if mail_type == "Guerrilla":
            email_address = "{}@guerrillamailblock.com".format(
                str(randrange(9_999_999_999))
            )
            session = GuerrillaMailSession(email_address=email_address)
            assert session.get_session_state()["email_address"] == email_address
            return EmailAccount(session=session, email_address=email_address)
        if mail_type == "nada":
            session = Nada(None, "getnada.com")
            return EmailAccount(session=session, email_address=str(session))
        raise NotImplementedError(f"Unsupported mail_type: {mail_type}")


class InviteLink:
    def __init__(self, account, email, token):
        self.account = account
        self.email = email
        self.token = token


class EmailUtils:
    INTERVAL = 3
    MAX_RETRIES = 15
    COLONY_EMAIL_ADDRESS = "noreply@quali.com"

    @staticmethod
    def wait_for_new_account_email(email_account: EmailAccount) -> EmailMessage:
        results = EmailUtils._wait_for(
            email_account=email_account,
            email_type="for new account",
            body_filter=lambda body: "Jump right in to get started" in body,
        )
        return results[0]

    @staticmethod
    def wait_for_new_user_email(email_account: EmailAccount, first_name: str) -> EmailMessage:
        results = EmailUtils._wait_for(
            email_account=email_account,
            email_type="for new users",
            body_filter=lambda body: "Congratulations {}! on joining".format(first_name)
            in body,
        )
        return results[0]

    @staticmethod
    def wait_for_invite_email(email_account: EmailAccount) -> InviteLink:
        results = EmailUtils._wait_for(
            email_account=email_account,
            email_type="to invite users",
            body_filter=lambda body: "You’ve been invited" in body,
        )

        var = re.search(
            '"http.*/register?email=(.+)&account=(.+)&token=(.+)"', results[0].body
        )
        return InviteLink(email=var.group(1), account=var.group(2), token=var.group(3))

    @staticmethod
    def wait_for_active_sandbox_email(email_account: EmailAccount, sandbox_name: str) -> EmailMessage:
        results = EmailUtils._wait_for(
            email_account=email_account,
            email_type="saying the sandbox is active",
            body_filter=lambda body: f"About {sandbox_name}:"
            and "is active and ready for use" in body,
        )
        return results[0]

    @staticmethod
    def _wait_for(email_account: EmailAccount, email_type: str, body_filter: Callable) -> List[EmailMessage]:
        all_emails = []
        for current_attempt in range(0, EmailUtils.MAX_RETRIES):
            all_emails = email_account.get_emails()
            colony_emails = [
                e for e in all_emails if e.sender == EmailUtils.COLONY_EMAIL_ADDRESS
            ]
            wanted_emails = [e for e in colony_emails if body_filter(e.body)]
            found = len(wanted_emails) > 0

            from gateways.utils import Utils
            print(
                "[{TIMESTAMP} utc] Attempt {ATTEMPT}/{MAX_ATTEMPTS} {CHECK_RESULT} - "
                "{EMAIL} got {ALL_LEN} emails, "
                "{COLONY_LEN} from colony, {WANTED_LEN} {TYPE}".format(
                    TIMESTAMP=Utils.get_current_time_utc().strftime("%X"),
                    ATTEMPT=current_attempt,
                    MAX_ATTEMPTS=EmailUtils.MAX_RETRIES,
                    CHECK_RESULT="succeeded" if found else "failed",
                    EMAIL=email_account.email_address,
                    ALL_LEN=len(all_emails),
                    COLONY_LEN=len(colony_emails),
                    WANTED_LEN=len(wanted_emails),
                    TYPE=email_type,
                )
            )
            if found:
                return wanted_emails

            time.sleep(EmailUtils.INTERVAL)

        if all_emails:
            EmailUtils._write_email_to_files(all_emails)
            print(
                "Writing all {} emails to filesystem (see artifacts)".format(
                    len(all_emails)
                )
            )
        else:
            print("No email is written to filesystems")

        raise Exception(
            "Timeout: {MAX_ATTEMPTS} attempts were exhausted waiting for email {TYPE}".format(
                MAX_ATTEMPTS=EmailUtils.MAX_RETRIES, TYPE=email_type
            )
        )

    @staticmethod
    def _write_email_to_files(emails: List[EmailMessage]):
        for i, email in enumerate(emails):
            file_name = f"email-{i}.txt"
            content = (
                f"Date:    {email.datetime}\n"
                f"Subject: {email.subject}\n"
                f"Sender:  {email.sender}\n"
                f"Body:    {email.body}"
            )
            with open(os.path.join(ArtifactPublisher.current_test_path, file_name), "w") as text_file:
                print(content, file=text_file)



from gateways.apis.api_base_class import ApiBase


class ApiSpaces(ApiBase):
    def spaces(self):
        return self.build_route("spaces")

    def space_by_name(self, space_name: str):
        return self.build_route("spaces/{space_name}".format(**locals()))

    def get_images(self, space_name: str, cloud_account_name: str, region_id: str):
        return self.build_route("spaces/{space_name}/cloud_accounts/{cloud_account_name}/regions/{region_id}/private_images"
                                .format(**locals()))

    def space_users(self, space_name: str):
        """Astronauts"""
        return self.build_route("spaces/{space_name}/users".format(**locals()))

    def space_repositories(self, space_name: str):
        return self.build_route("spaces/{space_name}/repositories".format(**locals()))

    def space_cloud_accounts(self, space_name: str):
        return self.build_route("spaces/{space_name}/cloud_accounts".format(**locals()))

    def space_cloud_account(self, space_name: str, cloud_account: str):
        return self.build_route(
            "spaces/{space_name}/cloud_accounts/{cloud_account}".format(**locals())
        )

    def space_github_repositories(self, space_name: str):
        return self.build_route(
            "spaces/{space_name}/repositories/github".format(**locals())
        )

    def space_bitbucket_repositories(self, space_name: str):
        return self.build_route(
            "spaces/{space_name}/repositories/bitbucket".format(**locals())
        )

    def user_permissions(self, space_name: str):
        return self.build_route("spaces/{space_name}/user_permissions".format(**locals()))

    def space_user(self, space_name: str, user_email: str):
        return self.build_route("spaces/{space_name}/users/{user_email}".format(**locals()))

    def user_space_role(self, space_name: str, user_email: str, space_role_name: str):
        return self.build_route(
            "spaces/{space_name}/users/{user_email}/space_role?value={space_role_name}".format(**locals())
        )

    def space_cloud_account_regions(self, space_name: str, cloud_account_name: str):
        return self.build_route(
            "spaces/{space_name}/cloud_accounts/{cloud_account_name}/regions".format(**locals())
        )



import json
import os
import urllib.parse
from typing import List

from time import sleep

import requests

from requests import Response

from gateways.common.cs18_api_classes import Commit


class GithubGateway:
    def __init__(self, user: str = "QualiNext", repo: str = "cs18-space-testing"):
        token = os.environ.get("GITHUB_ACCESS_TOKEN", "No GITHUB_ACCESS_TOKEN defined.")
        self.session = requests.Session()
        self.session.headers.update({"Authorization": "token " + token})
        self.session.headers.update({"Accept": "application/vnd.github.v3.raw"})
        self.base_api_url = "https://api.github.com/repos/{user}/{repo}".format(
            user=user, repo=repo
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()

    def close(self):
        self.session.close()

    def get_file(self, file_path, branch=None, tag=None, commit_id=None) -> str:
        url = "{github_api_url}/contents/{file_path}?ref={v}".format(
            github_api_url=self.base_api_url,
            file_path=file_path,
            v=branch or tag or commit_id,
        )
        return self.http_get(url=url).text

    def get_commit_history(self, file_path, branch="master") -> List[Commit]:
        url = "{github_api_url}/commits?sha={branch}&path={path}".format(
            github_api_url=self.base_api_url,
            branch=branch,
            path=urllib.parse.quote(file_path, safe=""),
        )
        response = self.http_get(url=url)
        items = json.loads(response.text)
        commits = [Commit(x) for x in items]
        sorted(commits, key=lambda x: x.date)
        return commits

    def http_get(self, url: str) -> Response:
        response = None
        attempt = 0
        while attempt < 5:
            response = self.session.get(url=url)
            if response.ok:
                return response
            attempt += 1
            print("Response from '{0}' returned {1}".format(url, response.status_code))
            sleep(1)
        response.raise_for_status()



from typing import List, Optional
from datetime import datetime
import dateutil.parser
from gateways.common.cs18_api_classes import CloudComputeService, AccessLink, CloudAccountCostResponse
from gateways.common.cs18_api_errors import SandboxError, Error
from gateways.common.cs18_api_responses import (
    SpaceAdmin,
    SpaceCloudAccount,
    GetSpacesResponse,
    CatalogForGetResponse,
    CatalogForGetAllResponse,
    BlueprintResponse,
    CloudAccountResponse,
    BlueprintInputResponse,
    SandboxResponse,
    ProductionBlueResponse, ProductionGreenResponse,
    SandboxOwnerResponse,
    SandboxResponseLean,
    ProductionResponseLean,
    CloudResponse,
    ApplicationResponse,
    ApplicationResponseLean,
    AppInstanceResponse,
    HyperlinkResponse,
    BlueprintApplicationResponse,
    BlueprintRegionResponse,
    BlueprintCloudResponse,
    TokenResponse,
    CatalogApplicationResponse,
    UserPermittedToSpaceResponse,
    SourceControlDetailsResponse,
    BlueprintFileResponse,
    UserInvitationResponse,
    RepositoryResponse,
    InputResponse,
    RoleListItemResponse,
    AccountStatusResponse,
    ArtifactRepoCreatedUser,
    AccountStatusUserDetails,
    ArtifactRepoResponse,
    UserForAllUsersResponse,
    ServiceResponse,
    BlueprintServiceResponse,
    LaunchingProgress, LaunchingProgressStep, CloudAccountInSpaceResponse, CloudAccountConnectivityResponse,
    CloudAccountInfraSetResponse, CloudAccountSubnetsResponse, RegionResponse, GetImagesResponse,
    VerifyCloudProviderResponse, VerifyCloudProviderResourceResponse)


class Converters:
    """
    Converts provided dictionaries into requested Response objects
    """

    @staticmethod
    def create_spaces_response(space: dict):
        admins = [
            SpaceAdmin(
                first_name=admin["first_name"],
                last_name=admin["last_name"],
                account_role=admin["account_role"],
            )
            for admin in space["admins"]
        ]
        space_cloud_accounts = [
            SpaceCloudAccount(
                name=cloud_account["name"], cloud_account_type=cloud_account["type"]
            )
            for cloud_account in space["cloud_accounts"]
        ]
        return GetSpacesResponse(
            name=space["name"],
            admins=admins,
            users_count=space["users_count"],
            cloud_accounts=space_cloud_accounts,
        )

    @staticmethod
    def create_catalog_blueprint_full_response(catalog: dict) -> CatalogForGetResponse:
        """
        Returns CatalogForGetResponse based on provided dictionary
        :param catalog: dict
        :return: CatalogForGetResponse
        """
        return CatalogForGetResponse(
            name=catalog["blueprint_name"],
            description=catalog["description"],
            url=catalog["url"],
            last_modified=catalog["last_modified"],
            modified_by=catalog["modified_by"],
            clouds=catalog["clouds"] if catalog["clouds"] else [],
            applications=[
                Converters._create_catalog_app_response(app)
                for app in catalog["applications"]
            ]
            if catalog["applications"]
            else [],
            links=[Converters._create_link_response(link) for link in catalog["links"]]
            if catalog["links"]
            else [],
            artifacts=catalog.get("artifacts", {}),
            errors=catalog.get("errors"),
        )

    @staticmethod
    def create_catalog_blueprint_response(catalog: dict) -> CatalogForGetAllResponse:
        """
        Returns CatalogForGetAllResponse based on provided dictionary
        :param catalog: dict
        :return: CatalogForGetAllResponse
        """
        return CatalogForGetAllResponse(
            name=catalog["blueprint_name"],
            description=catalog["description"],
            url=catalog["url"],
            last_modified=catalog["last_modified"],
            modified_by=catalog["modified_by"],
            clouds=catalog["clouds"] if catalog["clouds"] else [],
            links=[Converters._create_link_response(link) for link in catalog["links"]]
            if catalog["links"]
            else [],
            artifacts=catalog.get("artifacts", {}),
            errors=catalog.get("errors"),
        )

    @staticmethod
    def create_artifact_repo_response(obj: dict) -> ArtifactRepoResponse:
        created_by_json = obj["created_by"]
        return ArtifactRepoResponse(
            obj["type"],
            dateutil.parser.parse(obj["created_date"]),
            obj["location"],
            ArtifactRepoCreatedUser(
                created_by_json["email"],
                created_by_json["first_name"],
                created_by_json["last_name"],
            ),
        )

    @staticmethod
    def create_blueprint_response(blueprint: dict) -> BlueprintResponse:
        """
        Returns BlueprintResponse based on provided dictionary
        :param blueprint: dict
        :return: BlueprintResponse
        """
        return BlueprintResponse(
            name=blueprint["blueprint_name"],
            description=blueprint["description"],
            url=blueprint["url"],
            enabled=blueprint["enabled"],
            last_modified=blueprint["last_modified"],
            is_sample=blueprint["is_sample"],
            modified_by=blueprint["modified_by"],
            clouds=[
                Converters._create_blueprint_cloud_response(cloud)
                for cloud in blueprint["clouds"]
            ]
            if blueprint["clouds"]
            else [],
            applications=[
                Converters._create_blueprint_app_response(app)
                for app in blueprint["applications"]
            ]
            if blueprint["applications"]
            else [],
            links=[
                Converters._create_link_response(link) for link in blueprint["links"]
            ],
            artifacts=blueprint.get("artifacts", {}),
            inputs=[
                Converters._create_blueprint_input_response(i)
                for i in blueprint.get("inputs", [])
            ],
            errors=[Error(e) for e in blueprint["errors"]],
            services=[
                Converters._create_blueprint_service_response(app)
                for app in blueprint["services"]
            ]
            if blueprint["services"]
            else []
        )

    @staticmethod
    def _create_cloud_account_cost_response(cloud_account: dict) -> CloudAccountCostResponse:
        cost: dict = cloud_account.get('cost')
        if not cost:
            return CloudAccountCostResponse(None)

        return CloudAccountCostResponse(
            Converters._parse_datetime(cost.get("last_update")))

    @staticmethod
    def _create_errors(cloud_account: dict) -> List[Error]:
        errors: List[dict] = cloud_account.get('errors', [])
        return [Error(e) for e in errors]

    @staticmethod
    def _create_compute_services_response(
            cloud_compute_services: List[dict]
    ) -> List[CloudComputeService]:
        compute_services = [
            CloudComputeService(
                name=compute_service.get("name", None),
                service_type=compute_service.get("type", None),
                created_date=Converters._parse_datetime(compute_service.get("created_date", None)),
                created_by=compute_service.get("created_by", None),
                user_defined=compute_service.get("user_defined", None),
            )
            for compute_service in cloud_compute_services
        ]
        return compute_services

    @staticmethod
    def create_cloud_account_response(cloud: dict) -> CloudAccountResponse:
        """
        Returns CloudAccountResponse based on provided dictionary
        :param cloud: dict
        :return: CloudAccountResponse
        """
        return CloudAccountResponse(
            name=cloud["name"],
            provider=cloud["type"],
            created_date=Converters._parse_datetime(cloud.get("created_date")),
            created_by=cloud["created_by"],
            spaces=cloud["spaces"],
            compute_services=Converters._create_compute_services_response(
                cloud.get("compute_services", None)
            ),
            cost=Converters._create_cloud_account_cost_response(cloud),
            errors=Converters._create_errors(cloud)
        )

    @staticmethod
    def create_cloud_account_in_space_response(cloud: dict) -> CloudAccountInSpaceResponse:
        """
        Returns CloudAccountResponse based on provided dictionary
        :param cloud: dict
        :return: CloudAccountResponse
        """
        return CloudAccountInSpaceResponse(
            name=cloud["name"],
            provider=cloud["type"],
            created_date=Converters._parse_datetime(cloud.get("created_date")),
            created_by=cloud["created_by"],
            internet_facing=cloud["internet_facing"],
            connectivity=Converters._create_cloud_account_connectivity_response(cloud.get("connectivity")),
            compute_services=Converters._create_compute_services_response(
                cloud.get("compute_services", None)
            ),
            cost=Converters._create_cloud_account_cost_response(cloud),
            errors=Converters._create_errors(cloud)
        )

    @staticmethod
    def _create_cloud_account_connectivity_response(conn: dict) -> Optional[CloudAccountConnectivityResponse]:
        if not conn:
            return None

        return CloudAccountConnectivityResponse(
            existing_infra=conn.get("existing_infra"),
            existing_infra_sets=[CloudAccountInfraSetResponse(
                region=infra_set["region"],
                virtual_network=infra_set["virtual_network"],
                subnets=CloudAccountSubnetsResponse(
                    gateway_subnet=infra_set["subnets"].get("gateway_subnet"),
                    management_subnet=infra_set["subnets"].get("management_subnet"),
                    application_subnets=infra_set["subnets"].get("application_subnets")
                ) if "subnets" in infra_set else None) for infra_set in conn.get("existing_infra_sets")]
            if "existing_infra_sets" in conn else None)

    @staticmethod
    def _create_blueprint_input_response(
            blueprint_input: dict
    ) -> BlueprintInputResponse:
        return BlueprintInputResponse(
            name=blueprint_input["name"],
            default_value=blueprint_input.get("default_value"),
            display_style=blueprint_input.get("display_style"),
            optional=blueprint_input.get("optional", False),
            description=blueprint_input.get("description"),
        )

    @staticmethod
    def create_sandbox_response(sandbox: dict) -> SandboxResponse:
        """
        Returns SandboxResponse based on provided dictionary
        :param sandbox: dict
        :return: SandboxResponse
        """
        return SandboxResponse(
            sandbox_id=sandbox["id"],
            blueprint=sandbox["blueprint_name"],
            applications=[
                Converters._create_app_response(service)
                for service in sandbox["applications"]
            ]
            if sandbox["applications"]
            else [],
            services=[
                Converters._create_service_response(s)
                for s in sandbox.get("services", [])
            ],
            inputs=Converters._create_sandbox_inputs(sandbox.get("inputs")),
            artifacts=sandbox.get("artifacts"),
            automation=sandbox["automation"],
            start_time=sandbox["deployment_start_time"],
            create_time=sandbox["create_time"],
            status_details=sandbox["status_details"],
            sandbox_status=sandbox["sandbox_status"],
            status_error_description=sandbox["status_error_description"],
            links=[Converters._create_link_response(link) for link in sandbox["links"]]
            if sandbox["links"]
            else [],
            owner=Converters._create_sandbox_owner(sandbox["owner"])
            if sandbox["owner"]
            else [],
            errors=[
                Converters._create_sandbox_error(error)
                for error in sandbox.get("errors", [])
            ],
            scheduled_end_time=dateutil.parser.parse(sandbox["scheduled_end_time"])
            if sandbox.get("scheduled_end_time", None) else None,
            end_time=dateutil.parser.parse(sandbox["end_time"])
            if sandbox.get("end_time", None) else None,
            debugging_service=sandbox.get('debugging_service'),
            launching_progress=Converters._create_launching_progress(sandbox.get('launching_progress'))
        )

    @staticmethod
    def create_production_blue_response(production: dict) -> ProductionBlueResponse:
        """
        Returns ProductionBlueResponse based on provided dictionary
        :param production: dict
        :return: ProductionBlueResponse
        """
        return ProductionBlueResponse(
            production_id=production["id"],
            sandbox_id=production["sandbox_id"],
            name=production["name"],
            owner=Converters._create_sandbox_owner(production["owner"]),
            blueprint_name=production["blueprint_name"],
            blueprint_description=production["blueprint_description"],
            build=production["build"],
            applications=[
                Converters._create_app_response(app)
                for app in production.get("applications", [])
            ],
            services=[
                Converters._create_service_response(s)
                for s in production.get("services", [])
            ],
            inputs=Converters._create_sandbox_inputs(production.get("inputs")),
            artifacts=production.get("artifacts"),
            deployment_start_time=Converters._parse_datetime(
                production.get("deployment_start_time")
            ),
            create_time=Converters._parse_datetime(production.get("create_time")),
            status_details=production["status_details"],
            production_status=production["production_status"],
            update_status=production["update_status"],
            links=[
                Converters._create_link_response(link)
                for link in production.get("links") or []
            ],
            errors=[
                Converters._create_sandbox_error(error)
                for error in production.get("errors") or []
            ],
            status_error_description=production["status_error_description"],
            has_green=production["has_green"],
            debugging_service=production.get('debugging_service'),
            launching_progress=Converters._create_launching_progress(production.get('launching_progress'))
        )

    @staticmethod
    def create_production_green_response(production: dict) -> ProductionGreenResponse:
        """
        Returns ProductionGreenResponse based on provided dictionary
        :param production: dict
        :return: ProductionGreenResponse
        """
        return ProductionGreenResponse(
            production_id=production["id"],
            sandbox_id=production["sandbox_id"],
            name=production["name"],
            owner=Converters._create_sandbox_owner(production["owner"]),
            blueprint_name=production["blueprint_name"],
            blueprint_description=production["blueprint_description"],
            build=production["build"],
            applications=[Converters._create_app_response(app) for app in production.get("applications", [])],
            services=[
                Converters._create_service_response(s)
                for s in production.get("services", [])
            ],
            inputs=Converters._create_sandbox_inputs(production.get("inputs")),
            artifacts=production.get("artifacts"),
            deployment_start_time=Converters._parse_datetime(production.get("deployment_start_time")),
            create_time=Converters._parse_datetime(production.get("create_time")),
            status_details=production["status_details"],
            production_status=production["production_status"],
            update_status=production["update_status"],
            links=[Converters._create_link_response(link) for link in production.get("links") or []],
            errors=[Converters._create_sandbox_error(error) for error in production.get("errors") or []],
            status_error_description=production["status_error_description"],
            exposure=production["exposure"],
            debugging_service=production.get('debugging_service'),
            launching_progress=Converters._create_launching_progress(production.get('launching_progress')))

    @staticmethod
    def _create_sandbox_owner(user: dict) -> Optional[SandboxOwnerResponse]:
        if not user:
            return None

        return SandboxOwnerResponse(
            first_name=user["first_name"],
            last_name=user["last_name"],
            email=user["email"],
            join_date=user["join_date"],
        )

    @staticmethod
    def _create_sandbox_error(sandbox_error: dict) -> SandboxError:
        return SandboxError(
            time=sandbox_error["time"],
            code=sandbox_error["code"],
            message=sandbox_error["message"],
        )

    @staticmethod
    def create_sandbox_response_list_item(sandbox: dict) -> SandboxResponseLean:
        """
        Returns SandboxResponseLean based on provided dictionary
        :param sandbox: dict
        :return: SandboxResponseLean
        """
        return SandboxResponseLean(
            sandbox_id=sandbox["id"],
            blueprint=sandbox["blueprint_name"],
            applications=[
                Converters._create_application_response_lean(application)
                for application in sandbox["applications"]
            ]
            if sandbox["applications"]
            else [],
            services=[
                Converters._create_service_response(s)
                for s in sandbox.get("services", [])
            ],
            automation=sandbox["automation"],
            start_time=sandbox["deployment_start_time"],
            create_time=sandbox["create_time"],
            status_details=sandbox["status_details"],
            sandbox_status=sandbox["sandbox_status"],
            status_error_description=sandbox["status_error_description"],
            links=[Converters._create_link_response(link) for link in sandbox["links"]]
            if sandbox["links"]
            else [],
            owner=Converters._create_sandbox_owner(sandbox["owner"])
            if sandbox["owner"]
            else [],
            name=sandbox["name"],
            errors=[
                Converters._create_sandbox_error(error)
                for error in sandbox.get("errors", [])
            ],
            scheduled_end_time=dateutil.parser.parse(sandbox["scheduled_end_time"]) if sandbox.get(
                "scheduled_end_time") else None,
            end_time=dateutil.parser.parse(sandbox["end_time"]) if sandbox.get("end_time") else None,
            debugging_service=sandbox.get("debugging_service")
        )

    @staticmethod
    def create_production_response_list_item(
            production: dict
    ) -> ProductionResponseLean:
        """
        Returns ProductionResponseLean based on provided dictionary
        :param production: dict
        :return: ProductionResponseLean
        """
        return ProductionResponseLean(
            production_id=production["id"],
            name=production["name"],
            blueprint_description=production["blueprint_description"],
            blue_status=production["blue_status"],
            green_status=production["green_status"],
            update_strategy=production["update_strategy"],
            update_status=production["update_status"],
            create_time=Converters._parse_datetime(production.get("create_time")),
            deployment_start_time=Converters._parse_datetime(production.get("deployment_start_time")),
            blue_applications=[Converters._create_application_response_lean(x)
                               for x in production.get("blue_applications") or {}],
            blue_services=[Converters._create_service_response(s) for s in production.get("services", [])]
        )

    @staticmethod
    def _create_cloud_response(service: dict) -> CloudResponse:
        cloud_account_name = service.get("cloud_account_name")
        region_id = ""
        region_name = ""
        if "region" in service:
            region_id = service.get("region").get("id")
            region_name = service.get("region").get("name")
        return CloudResponse(
            cloud_account_name=cloud_account_name,
            region_id=region_id,
            region_name=region_name
        )

    @staticmethod
    def _create_app_response(application: dict) -> ApplicationResponse:
        return ApplicationResponse(
            name=application["name"],
            shortcuts=application["shortcuts"],
            status=application["status"],
            instances=[
                Converters._create_instance_response(instance)
                for instance in application["instances"]
            ],
            private_address=application["private_address"],
            public_address=application["public_address"],
            internal_ports=application["internal_ports"],
            external_ports=application["external_ports"],
            internal_dns=application["internal_dns"],
            image_name=application["image_name"],
            cloud=Converters._create_cloud_response(application.get("cloud")),
            icon=application.get("icon")

            if application.get("cloud")
            else None,
        )

    @staticmethod
    def _create_application_response_lean(application: dict) -> ApplicationResponseLean:
        return ApplicationResponseLean(
            name=application["name"],
            status=application["status"],
            cloud=Converters._create_cloud_response(application.get("cloud"))
            if application.get("cloud")
            else None,
        )

    @staticmethod
    def _create_instance_response(instance: dict) -> AppInstanceResponse:
        access_links = None
        if "access_links" in instance and instance["access_links"] is not None:
            access_links = [
                Converters._create_access_link(al) for al in instance["access_links"]
            ]

        return AppInstanceResponse(
            private_ip=instance["private_ip"],
            infra_id=instance["infra_id"],
            instance_type=instance["instance_type"],
            memory_limit=instance["memory_limit"],
            cpu_limit=instance["cpu_limit"],
            deployment_status=instance["status"],
            access_links=access_links,
            links=[Converters._create_link_response(link) for link in instance["links"]]
            if instance["links"]
            else [],
        )

    @staticmethod
    def _create_access_link(access_link_json: dict) -> AccessLink:
        return AccessLink(
            protocol=access_link_json["protocol"], link=access_link_json["link"]
        )

    @staticmethod
    def _create_link_response(link: dict) -> HyperlinkResponse:
        return HyperlinkResponse(
            rel=link["rel"], href=link["href"], method=link["method"]
        )

    @staticmethod
    def _create_blueprint_app_response(app: dict) -> BlueprintApplicationResponse:

        return BlueprintApplicationResponse(name=app["name"], version=app["version"])

    @staticmethod
    def _create_blueprint_service_response(service: dict) -> BlueprintServiceResponse:

        return BlueprintServiceResponse(name=service["name"], service_type=service["type"])

    @staticmethod
    def _create_blueprint_cloud_region_response(
            region: dict
    ) -> BlueprintRegionResponse:
        if not region:
            return BlueprintRegionResponse("", "")
        return BlueprintRegionResponse(
            region_id=region.get("id", ""), name=region.get("name", "")
        )

    @staticmethod
    def _create_blueprint_cloud_response(cloud):
        if isinstance(cloud, str):
            return cloud
        return BlueprintCloudResponse(
            cloud_account_name=cloud.get("cloud_account_name"),
            provider=cloud.get("provider"),
            region=Converters._create_blueprint_cloud_region_response(
                cloud.get("region")
            ),
        )

    @staticmethod
    def create_token_response(link: dict) -> TokenResponse:
        """
        Returns TokenResponse based on provided dictionary
        :param link: dict
        :return: TokenResponse
        """
        return TokenResponse(
            access_token=link["access_token"],
            refresh_token=link["refresh_token"],
            token_type=link["token_type"],
            expires_in=link["expires_in"],
        )

    @staticmethod
    def _create_catalog_app_response(app: dict) -> CatalogApplicationResponse:
        return CatalogApplicationResponse(name=app["name"], version=app["version"])

    @staticmethod
    def create_user_permitted_to_space_response(
            user: dict
    ) -> UserPermittedToSpaceResponse:
        """
        Returns UserPermittedToSpaceResponse based on provided dictionary
        :param user: dict
        :return: UserPermittedToSpaceResponse
        """
        return UserPermittedToSpaceResponse(
            first_name=user["first_name"],
            last_name=user["last_name"],
            email=user["email"],
            join_date=user["join_date"],
            space_role=user["space_role"],
            account_role=user["account_role"],
            has_access_to_all_spaces=user["has_access_to_all_spaces"]
        )

    @staticmethod
    def create_user_for_all_users_response(
            user: dict
    ) -> UserForAllUsersResponse:
        """
        Returns UserForAllUsersResponse based on provided dictionary
        :param user: dict
        :return: UserForAllUsersResponse
        """
        return UserForAllUsersResponse(
            email=user["email"],
            first_name=user["first_name"],
            last_name=user["last_name"],
            join_date=user["join_date"],
            account_role=user["account_role"],
            has_access_to_all_spaces=user["has_access_to_all_spaces"],
            member_in_spaces=user["member_in_spaces"]
        )

    @staticmethod
    def _create_source_control_details_response(
            source_control: dict
    ) -> SourceControlDetailsResponse:
        return SourceControlDetailsResponse(
            branch_name=source_control["branch_name"],
            latest_commit_id=source_control["latest_commit_id"],
            changed_by=source_control["changed_by"],
            changed_date=dateutil.parser.parse(source_control["change_date"]),
        )

    @staticmethod
    def create_blueprint_file_response(file: dict) -> BlueprintFileResponse:
        """
        Returns BlueprintFileResponse based on provided dictionary
        :param file: dict
        :return: BlueprintFileResponse
        """
        return BlueprintFileResponse(
            url=file["url"],
            kind=file["kind"],
            content=file["content"],
            source_control_details=Converters._create_source_control_details_response(
                file["source_control_details"]
            ),
            icon=file.get("icon"),
        )

    @staticmethod
    def create_user_invitation_response(invite: dict) -> UserInvitationResponse:
        """
        Returns UserInvitationResponse based on provided dictionary
        :param invite: dict
        :return: UserInvitationResponse
        """
        return UserInvitationResponse(
            email=invite["email"],
            send_date=invite["send_date"],
            account_role=invite["account_role"],
            space_name=invite["space_name"],
            space_role=invite["space_role"],
        )

    @staticmethod
    def create_repository_response(repo: dict) -> RepositoryResponse:
        """
        Returns RepositoryResponse based on provided dictionary
        :param repo: dict
        :return: RepositoryResponse
        """
        return RepositoryResponse(
            repository_url=repo["repository_url"],
            errors=[Error(e) for e in repo["errors"]],
        )

    @staticmethod
    def create_role_list_item_response(role: dict) -> RoleListItemResponse:
        return RoleListItemResponse(name=role["name"])

    @staticmethod
    def _create_sandbox_inputs(response_inputs: dict) -> List[InputResponse]:
        inputs = []
        if response_inputs:
            for request_input in response_inputs:
                inputs.append(
                    InputResponse(
                        name=request_input["name"],
                        value=request_input["value"],
                        description=request_input["description"],
                        display_style=request_input["display_style"],
                        optional=request_input.get("optional", False),
                    )
                )
        return inputs

    @staticmethod
    def create_account_status_response(obj: dict):
        return AccountStatusResponse(
            obj["onboarding_status"],
            Converters._parse_datetime(obj["license_expiration_date"]),
            obj["ignore_expiration_date"],
            obj["suspended_reason"],
            obj["salesforce_account_id"],
            obj["plan_id"],
            AccountStatusUserDetails(obj["user_details"]),
        )

    @staticmethod
    def _parse_datetime(value: str) -> datetime:
        if value:
            return dateutil.parser.parse(value)
        return None

    @staticmethod
    def _create_service_response(service: dict) -> ServiceResponse:
        return ServiceResponse(
            name=service["name"],
            status=service["status"],
            service_type=service["type"]
        )

    @staticmethod
    def _create_launching_progress(progress: dict) -> Optional[LaunchingProgress]:
        if not progress:
            return None
        return LaunchingProgress(
            creating_infrastructure=Converters._create_launching_progress_step(progress.get('creating_infrastructure')),
            preparing_artifacts=Converters._create_launching_progress_step(progress.get('preparing_artifacts')),
            deploying_applications=Converters._create_launching_progress_step(progress.get('deploying_applications')),
            verifying_environment=Converters._create_launching_progress_step(progress.get('verifying_environment')))

    @staticmethod
    def _create_launching_progress_step(step: dict) -> LaunchingProgressStep:
        return LaunchingProgressStep(
            total=step['total'],
            succeeded=step['succeeded'],
            failed=step['failed'],
            status=step['status'])

    @staticmethod
    def create_space_cloud_account_regions_response(region: dict) -> RegionResponse:
        return RegionResponse(region["id"], region["name"])

    @staticmethod
    def get_image_response(images: dict):
        return GetImagesResponse(images['image_id'], images['image_name'])

    @staticmethod
    def _create_resources_summary_response(resources_summary: dict) -> List[VerifyCloudProviderResourceResponse]:
        resources = []
        if resources_summary:
            for resource in resources_summary:
                resources.append(
                    VerifyCloudProviderResourceResponse(
                        resource_type=resource["resource_type"],
                        resource_name=resource["resource_name"],
                        resource_id=resource["resource_id"],
                        resource_status=resource["resource_status"],
                        resource_valid=resource["resource_valid"]
                    )
                )
        return resources

    @staticmethod
    def create_verify_response(verify: dict) -> VerifyCloudProviderResponse:
        return VerifyCloudProviderResponse(
            success=verify["success"],
            resources_summary=Converters._create_resources_summary_response(verify["resources_summary"]),
        )
